import { Directive, Injector, OnDestroy } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { Subscription } from 'rxjs';
import { ColorService } from '../services/vis/color.service';
import { StaticInjectorService } from './static-injector';

/**
 *  Mark the page as an angular directive.
 */
@Directive()
export abstract class AbstractPage implements OnDestroy {
  /**
   *  Subscription that - on change - emits the name of the new color scheme
   */
  colorSelectionChanged: Subscription;

  /**
   *  Holds the id_tag that represents a user
   */
  user: string;

  /**
   *  Holds the current color scheme
   */
  scheme = '';

  /**
   *  Indicates whether the components data has been loaded completely
   *  -> acts as a state for loading screens
   */
  componentStillLoading = true;

  /**
   *  Route is used to extract query parameters
   */
  protected route: ActivatedRoute;
  /**
   *  ColorService for subscribing to color scheme changes
   */
  protected colorService: ColorService;

  /**
   *  Set up common functionality thats used in most extending classes
   */
  protected constructor() {
    // inject dependencies so that you don't have to include them
    // in every super() call in every extending class
    const injector: Injector = StaticInjectorService.getInjector();
    this.route = injector.get(ActivatedRoute);
    this.colorService = injector.get(ColorService);

    // retrieve user id from url, set user 1 as default
    this.route.queryParams.subscribe((params) => {
      if (params.user === undefined) {
        this.user = '1';
      } else {
        this.user = params.user;
      }
    });

    // set the default color scheme
    this.scheme = this.colorService.barCharts.blue;

    // subscription that updates the color scheme on change
    this.colorSelectionChanged = this.colorService.changedColorSelection.subscribe((value) => {
      if (value) {
        this.scheme = this.colorService.barCharts[value];
        this.callAPIs();
      }
    });
  }

  /**
   *  Unsubscribe from all subscriptions once the page is destroyed
   */
  ngOnDestroy() {
    this.colorSelectionChanged.unsubscribe();
  }

  /**
   *  Abstract function that needs to be implemented by all extending classes
   *  Responsible for fetching the components data from the backend
   */
  abstract callAPIs(): void;
}
