import { Overlay, OverlayRef } from '@angular/cdk/overlay';
import { TemplatePortal } from '@angular/cdk/portal';
import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  OnDestroy,
  OnInit,
  TemplateRef,
  ViewChild,
} from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { default as moment } from 'moment';
import { Subject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';
import { DashboardElementDirective } from './directives/dashboard-element.directive';
import { IdleService } from './services/idle.service';
import { OneUpApiService } from './services/one-up-api.service';

/**
 *  The App's root component which defines app-wide behavior/functionality
 */
@Component({
  selector: 'app-root',
  templateUrl: './app.component.html',
  styleUrls: ['./app.component.css'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class AppComponent implements OnInit, OnDestroy {
  /**
   *  TemplateRef to the overlay template that's shown when the user is idle for 15min
   */
  @ViewChild('idleOverlay') idleOverlay: TemplateRef<unknown>;

  /**
   *  dbElement.viewContainerRef is used for creating the templatePortal
   */
  @ViewChild(DashboardElementDirective, { static: true }) dbElement: DashboardElementDirective;

  /**
   *  The actual overlay reference, that is used to attach and detach the overlay
   */
  overlayRef: OverlayRef;

  /**
   *  Portal that overlayRef attaches to
   */
  templatePortal: TemplatePortal<any>;

  /**
   *  Indicates whether the "user is idle"-overlay is shown
   */
  idleOverlayOpen = false;

  /**
   *  Observable which is indicates when the idleService
   *  should stop looking for user activity
   */
  private readonly destroyed$: Subject<void> = new Subject();

  /**
   *  Holds the id_tag that represents a user
   */
  private user: string;

  /**
   *  Inject the necessary services that are used in this component,
   *  sets up the idleStream, which monitors whether a user is active or idle
   */
  constructor(
    private overlay: Overlay,
    private oneUpApiService: OneUpApiService,
    private route: ActivatedRoute,
    private readonly changeDetectorRef: ChangeDetectorRef,
    private readonly idleService: IdleService
  ) {
    moment.locale('de');

    const seconds = 900; // 900 seconds = 15 min

    this.idleService
      .getIdleStream(seconds * 1000)
      .pipe(takeUntil(this.destroyed$))
      .subscribe((idle: boolean): void => {
        if (idle) {
          if (!this.idleOverlayOpen) {
            this.templatePortal = new TemplatePortal(
              this.idleOverlay,
              this.dbElement.viewContainerRef
            );
            const positionStrategy = this.overlay
              .position()
              .global()
              .height('300px')
              .centerHorizontally()
              .centerVertically();
            this.overlayRef = this.overlay.create({
              positionStrategy,
              maxHeight: '90vh',
              hasBackdrop: true,
            });
            this.idleOverlayOpen = true;
            this.logIdleStatus('userWentIdle');
            this.overlayRef.attach(this.templatePortal);
          }
        }
      });
  }

  /**
   *  Set the correct user id_tag when provided in the query
   */
  ngOnInit() {
    this.route.queryParams.subscribe((params) => {
      if (params.user === undefined) {
        this.user = '1';
      } else {
        this.user = params.user;
      }
    });
  }

  /**
   *  Logs the user activity status ('userWentIdle' / 'userReturned') to the backend log
   */
  logIdleStatus(activity: string) {
    this.oneUpApiService.postUserActivity(this.user, activity).subscribe(() => {});
  }

  /**
   *  Closes the idle-overlay and marks the user as "returned"
   */
  userIsBack() {
    this.overlayRef.detach();
    this.idleOverlayOpen = false;
    this.logIdleStatus('userReturned');
  }

  /**
   *  Stops watching for user activity stream when component is destroyed
   */
  ngOnDestroy(): void {
    this.destroyed$.next();
    this.destroyed$.complete();
  }
}
