import { Overlay, OverlayRef } from '@angular/cdk/overlay';
import { TemplatePortal } from '@angular/cdk/portal';
import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnDestroy,
  OnInit,
  TemplateRef,
  ViewChild,
} from '@angular/core';
import { FormControl } from '@angular/forms';
import { PageEvent } from '@angular/material/paginator';
import { ActivatedRoute } from '@angular/router';
import { Observable, Subscription } from 'rxjs';
import { map, startWith } from 'rxjs/operators';
import { DashboardElementDirective } from '../../directives/dashboard-element.directive';
import { TagsOverview, UserTopic } from '../../interfaces/one-up-api-response';
import { HighlightService } from '../../services/highlight.service';
import { HttpRequestCache } from '../../services/http-request-cache.service';
import { OneUpApiService } from '../../services/one-up-api.service';
import { ColorService } from '../../services/vis/color.service';
import { convertToClassName, filter } from '../../util';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-topic-selection-bar',
  templateUrl: './topic-selection-bar.component.html',
  styleUrls: ['./topic-selection-bar.component.css'],
})
export class TopicSelectionBarComponent implements AfterViewInit, OnDestroy, OnInit {
  /**
   *  Holds the id_tag that represents a user
   */
  @Input() user: string;

  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any>;

  /**
   *  Holds the id_tag that represents a user
   */
  @Input() disableCategoriesOverlay = false;

  /**
   *  ElementRef to the topic search bar input element
   */
  @ViewChild('topicInput') topicInput: ElementRef<HTMLInputElement>;

  /**
   *  ElementRef to the topic search bar overlay that suggests topics based on the input
   */
  @ViewChild('topicInputOverlay') topicInputOverlay: ElementRef<HTMLInputElement>;

  /**
   *  TemplateRef to the overlay template that lets users browser through
   *  and select new topics
   */
  @ViewChild('templatePortalContent') templatePortalContent: TemplateRef<unknown>;

  /**
   *  ng-template that is used to create a TemplatePortal, to which the overlay then attaches.
   */
  @ViewChild(DashboardElementDirective, { static: true }) dbElement: DashboardElementDirective;

  /**
   *  For every parent category: store how many topics are currently selected.
   */
  selectionPerParent = {};

  /**
   *  FormControl for the topic search bar
   */
  topicCtrl = new FormControl('');

  /**
   *  Holds all topics that match the users search query
   */
  filteredTopics: Observable<string[]>;

  /**
   *  Holds the currently shown topics in the accordion menu (overlay)
   *  when a category is expanded.
   */
  currentTagList = {};

  /**
   *  Holds the class name of the currently highlightedTopic (if any).
   */
  highlightedTopic = '';

  /**
   *  Holds all topics that the API returns
   */
  allTopics: string[] = [];

  /**
   *  Subscribes to the highlightedTopicChanged subscription to
   *  always receive the latest highlightedTopic
   */
  highlightSubscription: Subscription;

  /**
   *  Boolean indicator whether an overlay is already open.
   *  Prevents multiple overlays from being displayed on top of each other.
   */
  overlayIsOpen = false;

  /**
   *  The actual overlay reference, that is used to attach and detach the overlay.
   */
  overlayRef: OverlayRef;

  /**
   *  Portal that overlayRef attaches to.
   */
  templatePortal: TemplatePortal<any>;

  /**
   *  Holds all topics that the user has currently selected.
   */
  selectedTopics = [];

  /**
   *  Boolean indicator for every available topic, showing whether it is selected or not.
   *  e.g. this.topicSelection['Strafrecht'] = true;
   */
  topicSelection: any = {};

  /**
   *  Holds all available tags and its parent tags.
   */
  tagList = [{ name: 'Parent1', tags: [{ name: 'Test1' }, { name: 'Test2' }] }];

  /**
   *  Boolean that indicates whether the categories overlay should be displayed.
   */
  showCategories = false;

  /**
   *  Holds all visual parameters, in this case the color scheme
   */
  visOptions = {
    scheme: this.colorService.barCharts.blue,
  };

  /**
   *  Inject the necessary services that are used in this component
   */
  constructor(
    private oneUpApiService: OneUpApiService,
    private highlightService: HighlightService,
    private colorService: ColorService,
    private cache: HttpRequestCache,
    private overlay: Overlay,
    private route: ActivatedRoute,
    private cdRef: ChangeDetectorRef
  ) {
    this.filteredTopics = this.topicCtrl.valueChanges.pipe(
      startWith(null),
      map((topic: string | null) => filter(topic, this.allTopics, this.topicSelection))
    );
  }

  /**
   *  Loads all available topics and initializes corresponding data structures.
   *  Sets up a subscription to listen to changes regarding the highlightedTopic.
   */
  ngOnInit() {
    this.route.queryParams.subscribe((params) => {
      if (params.cat && params.cat > 0) {
        this.showCategories = true;
      }
    });

    this.topicSelectionChanged.subscribe((value: any) => {
      // when the topic selection changes while the overlay is open
      // this updates the topic selection in -underlying- instances of the topic-selection-bar
      if (!this.overlayIsOpen) {
        this.callAPIs(true);
      }
    });

    this.callAPIs();

    this.highlightSubscription = this.highlightService.highlightedTopicChanged.subscribe(
      (value) => {
        this.highlightedTopic = this.getClassName(value);
        this.cdRef.detectChanges();
      }
    );
  }

  /**
   *  Sets up overlay related variables, creates an overlay backdrop listener
   */
  ngAfterViewInit() {
    this.templatePortal = new TemplatePortal(
      this.templatePortalContent,
      this.dbElement.viewContainerRef
    );
    const scrollStrategy = this.overlay.scrollStrategies.block();

    this.overlayRef = this.overlay.create({
      scrollStrategy,
      maxHeight: '90vh',
      hasBackdrop: true,
    });

    this.overlayRef.backdropClick().subscribe(() => {
      if (this.selectedTopics.length > 0) {
        this.detachTemplate();
      }
    });

    if (this.showCategories) {
      this.attachOverlay();
    }
  }

  /**
   *  Unsubscribe from subscriptions when page is destroyed
   */
  ngOnDestroy() {
    this.highlightSubscription.unsubscribe();
  }

  /**
   *  Fetches all tags and the users favorite tags from the API
   */
  callAPIs(suppressOverlay?: boolean) {
    this.oneUpApiService.getTagsOverview().subscribe((response) => {
      this.initTagList(response);
      this.oneUpApiService
        .getUserTopics(this.user)
        .subscribe((response) => this.initSelectedList(response, suppressOverlay));
    });
  }

  /**
   *  Parses the getTagsOverview API response and initializes the tagList
   *  which holds all available topics.
   *
   *  @param response – of the getTagsOverview API
   */
  initTagList(response: TagsOverview[]) {
    this.tagList = response;
    this.selectionPerParent = {};
    this.allTopics = [];
    response.forEach((parent) => {
      this.setPaginatedList(parent, 0, 10);
      this.selectionPerParent[parent.name] = 0;
      parent.tags.forEach((tag) => {
        this.allTopics.push(tag.name);
        this.topicSelection[tag.name] = false;
      });
    });
  }

  /**
   *  Parses the getUserTopics API response and marks the users favorite topics as selected
   *
   *  @param response – of the getUserTopics API
   */
  initSelectedList(response: UserTopic[], suppressOverlay?: boolean) {
    if (response.length === 0 && !suppressOverlay) {
      this.attachOverlay();
      return;
    }

    if (response[0].name === 'No User') {
      this.showCategories = true;
      return;
    }

    response.forEach((topic) => {
      this.topicSelection[topic.name] = true;
    });

    this.getSelectedTopics();
    // set the value once to init filteredTopics
    this.topicCtrl.setValue(null);
  }

  /**
   * Responsible to show the selected elements in the overlay categories accordion menu.
   *
   * @param tagList - The tagList thats supposed to be paginated
   * @param begin - The beginning index of the portion of the tagList to show
   * @param end - The end index. This is exclusive of the element at the index 'end'
   */
  setPaginatedList(tagList: TagsOverview, begin: number, end: number) {
    this.currentTagList[tagList.name] = tagList.tags.slice(begin, end);
  }

  /**
   * Adds a topic to the users favorite/selected topics and sends the new list to the backend.
   * Then clears the input of the topic search bar, its form and the topic search bar overlay.
   *
   * @param event - Event of the mat-autocomplete overlay, once a users selects a new topic.
   */
  selected(event) {
    this.topicSelection[event.option.viewValue] = true;
    this.updateUserTags(event.option.viewValue);
    this.topicInput.nativeElement.value = '';
    if (this.topicInputOverlay) {
      this.topicInputOverlay.nativeElement.value = '';
    }
    this.topicCtrl.setValue(null);
  }

  /**
   * Removes a topic from the users favorite/selected topics.
   * Also removes the highlighting from a topic.
   * Sends the updated user favorite topics to the backend.
   * If there aren't anymore topics selected by the user, it shows the categories overlay.
   *
   * @param topic - Name of the topic to be deselected.
   */
  deselectTopic(topic: string) {
    this.highlightService.removeHighlighting();

    this.topicSelection[topic] = false;
    this.updateUserTags(topic);

    if (this.selectedTopics.length === 0) {
      this.activateOverlay();
    }
  }

  /**
   * Gets the users currently selected topics and sends/saves them to the backend.
   */
  updateUserTags(topic?: string) {
    this.getSelectedTopics();
    this.oneUpApiService.postUserTags(this.user, this.selectedTopics).subscribe(() => {
      this.cache.clear();
      this.topicSelectionChanged.emit(topic ? topic : null);
    });
  }

  /**
   *  Initializes the overlay attachment and refreshes the currently selected topics.
   */
  activateOverlay() {
    this.attachOverlay();
    this.getSelectedTopics();
  }

  /**
   * Closes the overlay and returns to the Dashboard
   */
  detachTemplate() {
    if (this.overlayRef && this.overlayRef.hasAttached()) {
      this.overlayRef.detach();
    }
    this.overlayIsOpen = false;
  }

  /**
   * Changes the currently shown tags of a tagList based on the new pageIndex
   */
  changedPaginator(event: PageEvent, parent) {
    const baseNumber = event.pageIndex * event.pageSize;
    this.currentTagList[parent.name] = parent.tags.slice(baseNumber, baseNumber + event.pageSize);
  }

  /**
   * Updates the users favorite/selected topics when a checkbox in the categories overlay changes.
   * Also removes the highlighting.
   */
  changedCheckbox() {
    this.highlightService.removeHighlighting();
    this.updateUserTags();
  }

  /**
   *  Get a topics CSS class name by remove all whitespace in the string.
   *
   *  @param topic – a topics name as string
   */
  getClassName(topic: string) {
    return convertToClassName(topic);
  }

  /**
   *  Highlights a specific topic (with a red border) after the user has clicked on it.
   *
   *  @param topic - a topics name as string
   */
  highlightTopic(topic: string) {
    this.highlightedTopic = this.getClassName(topic);

    if (topic) {
      this.highlightService.highlightGraphs(topic);
    }
  }

  /**
   *  Attaches the overlay to the DOM.
   */
  private attachOverlay() {
    if (!this.overlayIsOpen) {
      this.overlayIsOpen = true;
      this.overlayRef.attach(this.templatePortal);
    }
  }

  /**
   *  Returns an array of strings, containing all selected topics.
   */
  private getSelectedTopics(): string[] {
    // forget old array
    this.selectedTopics.length = 0;
    const hasSelectionPerParent = this.selectionPerParent !== undefined;
    if (hasSelectionPerParent) {
      for (const name of Object.keys(this.selectionPerParent)) {
        this.selectionPerParent[name] = 0;
      }
    }
    for (const key in this.topicSelection) {
      if (this.topicSelection[key]) {
        this.selectedTopics.push(key);
        if (hasSelectionPerParent) {
          for (const parentKey of this.tagList.keys()) {
            const parent = this.tagList[parentKey];
            if (parent.tags.filter((tag) => tag.name === key).length > 0) {
              this.selectionPerParent[parent.name] += 1;
            }
          }
        }
      }
    }
    return this.selectedTopics;
  }
}
