import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  OnDestroy,
  ViewChild,
} from '@angular/core';
import * as d3 from 'd3';
import { default as moment } from 'moment';
import { BsModalRef } from 'ngx-bootstrap/modal';
import { Subscription } from 'rxjs';
import { HighlightService } from '../services/highlight.service';
import { MediaOverviewType, OneUpApiService, TimespanType } from '../services/one-up-api.service';
import { PrService } from '../services/pr.service';
import { BarchartService } from '../services/vis/barchart.service';
import { CircleGraphService } from '../services/vis/circle-graph.service';
import { ColorService } from '../services/vis/color.service';
import { LineGraphService } from '../services/vis/line-graph.service';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-dashboard-elements-overlay',
  templateUrl: './dashboard-elements-overlay.modal.html',
  styleUrls: ['./dashboard-elements-overlay.modal.css', '../app.component.css'],
})
export class DashboardElementsOverlayComponent implements AfterViewInit, OnDestroy {
  /**
   *  ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('diagram') chartContainer: ElementRef;

  /**
   *  title of the component to be shown in the overlay
   */
  title: string;

  /**
   *  name of the component in the wiki
   */
  wikiName: string;

  /**
   *  holds different values that are passed over for diagram creation
   */
  visOptions: any;

  /**
   *  identifies the user by id
   */
  user: string;

  /**
   *  holds the name of the highlighted topic
   */
  highlightedTopic: string;

  /**
   *  holds the color of the highlighted topic
   */
  highlightedTopicColor: string;

  /**
   * subscribes to the highlightedTopicChanged event to update highlightedTopic
   */
  highlightSubscription: Subscription;

  /**
   *  value that represents the active tab
   *  for app-daterange-toggle-group or app-topic-toggle-group
   */
  activeTab: string;

  /**
   *  EventEmitter that fires when the user selected topics change
   */
  topicSelectionChanged: EventEmitter<any>;

  /**
   *  time period for which data is retrieved
   */
  selectedTimespan: TimespanType;

  /**
   *  indicator for the loading spinner
   */
  componentStillLoading = true;

  /**
   *  holds an array of the most recent articles
   */
  mostRecentArticles;

  /**
   *  subscribes to clicking events on the bars
   *  to then show the most recent articles for the selected bar
   */
  changeMediumDetailSubscription: Subscription;

  /**
   *  hot-media: holds the mediums name once its corresponding bar has been clicked
   *  and the list view of the mediums most recent articles is fetched
   */
  selectedMedium;

  /**
   *  inject the necessary services that are used in this component
   */
  constructor(
    /**
     *  Is accessed in the HTML to close the overlay
     */
    public bsModalRef: BsModalRef,
    private lineGraphService: LineGraphService,
    private oneUpApiService: OneUpApiService,
    private barchartService: BarchartService,
    private circleGraphService: CircleGraphService,
    private colorService: ColorService,
    private prService: PrService,
    private highlightService: HighlightService,
    private cdRef: ChangeDetectorRef
  ) {}

  /**
   *  Initializes the bar chart by
   *    - drawing the SVG
   *    - fetching the API
   *
   *  Subscribes to topicSelectionChanged,
   *  and defines the functionality when a change is received
   */
  ngAfterViewInit() {
    this.topicSelectionChanged.subscribe(() => this.callAPIs());
    this.visOptions.scheme = this.colorService.barCharts.blue;
    this.visOptions.height = 690;
    this.visOptions.width = 1500;
    this.createSVGs();
    this.callAPIs();

    this.highlightSubscription = this.highlightService.highlightedTopicChanged.subscribe(
      (value) => {
        this.activeTab = value ? '2' : '1';
        this.highlightedTopic = value ? value : '';
        this.callAPIs();
      }
    );
  }

  /**
   *  Unsubscribes from active subscriptions once the component is destroyed.
   *
   *  Hide all tooltips when the overlay is closed,
   *  prevents tooltips from getting stuck because mouseleave event doesn't fire.
   */
  ngOnDestroy() {
    d3.selectAll('#hoverTooltip').style('display', 'none');
    if (this.title === 'Medien-Übersicht') {
      this.changeMediumDetailSubscription.unsubscribe();
    }

    this.highlightSubscription.unsubscribe();
  }

  /**
   *  Modify visOptions for the diagram depending on the displayed component.
   *  Create the corresponding SVG.
   */
  createSVGs() {
    switch (this.title) {
      case 'Themenreichweite':
        this.visOptions.element = this.lineGraphService.createSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        break;
      case 'Artikelanzahl':
        this.visOptions.element = this.lineGraphService.createSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        break;
      case 'Regionale Themen':
        this.visOptions.height = 690;
        this.visOptions.margin = { top: 10, bottom: 40, left: 40, right: 10 };
        this.visOptions.scrollDirection = 'vertical';
        this.visOptions.xAxis = this.barchartService.createAxisSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        this.visOptions.element = this.barchartService.createBody(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        break;
      case 'Angesagte Themen im Vergleich':
        this.visOptions.height = 730;
        this.visOptions.margin = { top: 10, bottom: 40, left: 40, right: 10 };
        this.visOptions.scrollDirection = 'vertical';
        this.visOptions.xAxis = this.barchartService.createAxisSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        this.visOptions.element = this.barchartService.createBody(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        break;
      case 'Popularität':
        this.visOptions.height = 700;
        this.visOptions.margin = { top: 10, bottom: 20, left: 60, right: 10 };
        this.visOptions.scrollDirection = 'horizontal';
        this.visOptions.yAxis = this.barchartService.createAxisSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        this.visOptions.element = this.barchartService.createBody(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        break;
      case 'Öffentliche Wahrnehmung':
        this.visOptions.height = 745;
        this.visOptions.margin = { top: 0, bottom: 10, left: 10, right: 10 };
        this.visOptions.element = this.circleGraphService.createSVG(
          this.chartContainer.nativeElement,
          this.visOptions,
          true
        );
        break;
      case 'Medien-Übersicht':
        this.visOptions.height = 670;
        this.visOptions.width = 1550 * 0.5;
        this.visOptions.margin = { top: 10, bottom: 45, left: 0, right: 10 };
        this.visOptions.scrollDirection = 'vertical';
        this.visOptions.xAxis = this.barchartService.createAxisSVG(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        this.visOptions.element = this.barchartService.createBody(
          this.chartContainer.nativeElement,
          this.visOptions
        );
        this.changeMediumDetailSubscription =
          this.barchartService.changeMediumDetailEmitter.subscribe((value) => {
            if (value) {
              const obj = JSON.parse(value);
              const arr = this.getMediaOverviewTypeAndQuery();
              this.updateRecentArticles(arr[0], arr[1], obj.mediumName);
            }
          });
        break;
    }
  }

  /**
   *  Fetch data from the corresponding API
   */
  callAPIs() {
    this.componentStillLoading = true;
    switch (this.title) {
      case 'Themenreichweite':
        this.showTopicReach();
        break;
      case 'Artikelanzahl':
        this.showArticleCount();
        break;
      case 'Regionale Themen':
        this.showRegionalInfluence();
        break;
      case 'Angesagte Themen im Vergleich':
        this.showHotTopics();
        break;
      case 'Popularität':
        this.showTopicPopularity();
        break;
      case 'Öffentliche Wahrnehmung':
        this.showTopicCircle();
        break;
      case 'Medien-Übersicht':
        this.showHotMedia();
        break;
    }
  }

  /**
   *  Makes sure that changes are represented in the DOM
   */
  componentFinishedLoading() {
    this.componentStillLoading = false;
    this.cdRef.detectChanges();
  }

  /**
   *  Changes the toggle mode and calls the API to get the corresponding data.
   *
   *  @param event – holds the value of the selected mat-button-toggle
   */
  changeSelection(selectedTab) {
    this.activeTab = selectedTab;
    this.callAPIs();
  }

  /**
   *  Deselect a specific medium for hot-media and fetches
   *  recent articles for all media
   */
  deselectMedium() {
    this.selectedMedium = undefined;
    this.callAPIs();
  }

  /**
   *  Receives output from the timespan-selector component,
   *  sets the newly selected timespan and refreshes the data
   */
  timespanSelectionChanged(timespan) {
    this.selectedTimespan = TimespanType[timespan];
    this.callAPIs();
  }

  /**
   *  Depending on the activeTab, this functions build and calls the getTagsTimeview API
   *  and creates (or updates) the article count line graph
   */
  showTopicReach() {
    const arr = this.getMediaOverviewTypeAndQuery();
    const mediaOverviewType = arr[0];
    const query = arr[1];

    this.oneUpApiService.getTagsTimeview(mediaOverviewType, query).subscribe((response) => {
      this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
        this.lineGraphService.createUpdateLineGraph(this.visOptions, response, userFavoriteTopics);
        this.componentFinishedLoading();
      });
    });
  }

  /**
   *  Uses the same API as TopicReach, only difference is in visOptions.yColumnName
   */
  showArticleCount() {
    this.showTopicReach();
  }

  /**
   *  Fetches the getUserTagsWithArticleCount API and draws (or updates) the pyramid chart
   */
  showRegionalInfluence() {
    this.oneUpApiService
      .getUserTagsWithArticleCount(this.user, MediaOverviewType.allTags, this.selectedTimespan)
      .subscribe((response) => {
        this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
          this.barchartService.createUpdatePyramidBars(
            this.visOptions,
            response,
            userFavoriteTopics,
            true
          );
          this.componentFinishedLoading();
        });
      });
  }

  /**
   *  Fetches the getHotTopics API and draws the pyramid chart
   */
  showHotTopics() {
    if (this.activeTab === '0') {
      this.visOptions.xColumnLeftName = 'diff_last_week';
      this.visOptions.xColumnRightName = 'diff_last_week';
      this.visOptions.xTitle = 'Veröffentlichungen – Differenz zur Vorwoche';
    } else {
      this.visOptions.xColumnLeftName = 'diff_last_month';
      this.visOptions.xColumnRightName = 'diff_last_month';
      this.visOptions.xTitle = 'Veröffentlichungen – Differenz zum Vormonat';
    }

    this.oneUpApiService
      .getHotTopics(this.user, MediaOverviewType.allTags)
      .subscribe((response) => {
        this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
          this.barchartService.createUpdatePyramidBars(
            this.visOptions,
            response,
            userFavoriteTopics,
            true
          );
          this.componentFinishedLoading();
        });
      });
  }

  /**
   *  Fetches the getUserTagsWithArticleCount API and draws (or updates) the pyramid chart
   */
  showTopicPopularity() {
    this.oneUpApiService
      .getUserTagsWithArticleCount(this.user, MediaOverviewType.allTags, this.selectedTimespan)
      .subscribe((response) => {
        this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
          this.barchartService.createUpdateVerticalBars(
            this.visOptions,
            response,
            userFavoriteTopics,
            true
          );
          this.componentFinishedLoading();
        });
      });
  }

  /**
   *  Fetches data from the getTopics API and the getUserTopics API and
   *  uses that data to create the circleGraph diagram
   */
  showTopicCircle() {
    this.oneUpApiService.getTopics(this.user, MediaOverviewType.allTags).subscribe((response) => {
      if (response === null || response.length === 0) {
        response = this.prService.getFeedback();
      }

      this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
        this.circleGraphService.createChart(
          this.visOptions.element,
          this.visOptions.height,
          this.visOptions.width,
          response,
          this.visOptions.scheme,
          userFavoriteTopics,
          true
        );
        this.highlightService.removeHighlighting();
        this.componentFinishedLoading();
      });
    });
  }

  /**
   *  Calls the API to update the most popular media (getHotMedia)
   *  and the most recent articles (getRecentArticles)
   *  and creates (or updates) the corresponding view
   */
  showHotMedia(mediumName?) {
    const arr = this.getMediaOverviewTypeAndQuery();
    const mediaOverviewType = arr[0];
    const query = arr[1];

    this.oneUpApiService
      .getHotMedia(mediaOverviewType, this.selectedTimespan, query, true)
      .subscribe((response) => {
        this.visOptions.scheme =
          this.activeTab === '2'
            ? d3.scaleOrdinal().range([this.highlightedTopicColor])
            : this.colorService.barCharts.mediaOverview;
        this.visOptions.maxRange = response[0].article_count;
        this.barchartService.createUpdateHorizontalBars(this.visOptions, response, true);
        this.updateRecentArticles(mediaOverviewType, query, mediumName);
      });
  }

  /**
   *  Updates the most recent articles
   */
  updateRecentArticles(mediaOverviewType, query, mediumName) {
    this.componentStillLoading = true;
    this.selectedMedium = mediumName;
    this.oneUpApiService
      .getRecentArticles(mediaOverviewType, 50, query, mediumName ? mediumName : 'all')
      .subscribe((response) => {
        const data = JSON.parse(JSON.stringify(response));
        for (const article of data) {
          article.date_of_publication = moment(article.date_of_publication).format('DD. MMMM YYYY');
        }

        this.mostRecentArticles = data;
        this.componentFinishedLoading();
      });
  }

  /**
   *  Sets the mediaOverviewType and the query
   *  based on the selected app-topic-toggle-group tab
   */
  getMediaOverviewTypeAndQuery() {
    let mediaOverviewType;
    let query;
    switch (this.activeTab) {
      case '0':
        mediaOverviewType = MediaOverviewType.allTags;
        break;
      case '1':
        mediaOverviewType = MediaOverviewType.userTags;
        query = this.user;
        break;
      case '2':
        mediaOverviewType = MediaOverviewType.queryTag;
        query = this.highlightedTopic;
        break;
    }

    return [mediaOverviewType, query];
  }

  /**
   *  Logs activity to the backend when the user opens an article
   */
  logUserActivity(title: string) {
    this.oneUpApiService
      .postUserActivity(this.user, 'userOpenedArticle', {
        activity: 'userOpenedArticle',
        articleName: title,
      })
      .subscribe(() => {});
  }
}
