import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnInit,
  ViewChild,
} from '@angular/core';
import { BsModalService } from 'ngx-bootstrap/modal';
import { HighlightService } from 'src/app/services/highlight.service';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { DashboardElementsOverlayComponent } from '../../dashboard-elements-overlay/dashboard-elements-overlay.modal';
import { MediaOverviewType, OneUpApiService } from '../../services/one-up-api.service';
import { PrService } from '../../services/pr.service';
import { CircleGraphService } from '../../services/vis/circle-graph.service';

/**
 * defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-article-topic-circle',
  templateUrl: './article-topic-circle.component.html',
  styleUrls: ['./article-topic-circle.component.css', '../../app.component.css'],
})
export class ArticleTopicCircleComponent extends AbstractPage implements OnInit, AfterViewInit {
  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();
  /**
   * ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('chart') private chartContainer: ElementRef;

  /**
   * define the charts view margin
   */
  margin = { top: 10, bottom: 10, left: 10, right: 10 };
  /**
   * define the charts view width
   */
  width = 500 - this.margin.left - this.margin.right;
  /**
   * define the charts view height
   */
  height = 240 - this.margin.top - this.margin.bottom;

  /**
   * holds different values that are passed over for diagram creation
   */
  visOptions = {
    margin: this.margin,
    width: this.width,
    height: this.height,
    scheme: this.colorService.barCharts.blue,
    element: undefined,
  };

  /**
   * inject the necessary services that are used in this component
   */
  constructor(
    private prService: PrService,
    private circleGraphService: CircleGraphService,
    private highlightService: HighlightService,
    private oneUpApiService: OneUpApiService,
    private modalService: BsModalService,
    private cdRef: ChangeDetectorRef
  ) {
    super();
  }

  /**
   * On component initialization,
   *  - retrieve the user-ID from the query parameters
   *  - set the default color scheme
   *  - create listener, that refreshes the diagram when the topic selection changes
   *  - create listener, that refreshes the diagram when the color scheme changes
   */
  ngOnInit(): void {
    this.topicSelectionChanged.subscribe(() => this.callAPIs());
  }

  /**
   * Once the component is initialized, create an SVG that will later display the diagram.
   * Then, call the APIs to initialize and create the diagram.
   */
  ngAfterViewInit(): void {
    this.visOptions.element = this.circleGraphService.createSVG(
      this.chartContainer.nativeElement,
      this.visOptions
    );
    this.callAPIs();
  }

  /**
   * Passes data to the modalService and opens diagram in the DashboardElementsOverlayComponent
   */
  openOverlay() {
    const initialState = {
      title: 'Öffentliche Wahrnehmung',
      wikiName: 'Öffentliche-Wahrnehmung',
      user: this.user,
      visOptions: JSON.parse(JSON.stringify(this.visOptions)),
      topicSelectionChanged: this.topicSelectionChanged,
    };
    this.modalService.show(DashboardElementsOverlayComponent, { initialState });
  }

  /**
   * Sends a request to the getTopics API and the getUserTopics API and
   * uses that data to create the circleGraph diagram
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.visOptions.scheme = this.scheme;
    this.oneUpApiService.getTopics(this.user, MediaOverviewType.userTags).subscribe((response) => {
      if (response === null || response.length === 0) {
        response = this.prService.getFeedback();
      }

      this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
        this.circleGraphService.createChart(
          this.visOptions.element,
          this.height,
          this.width,
          response,
          this.visOptions.scheme,
          userFavoriteTopics
        );
        this.highlightService.removeHighlighting();
        this.componentStillLoading = false;
        this.cdRef.detectChanges();
      });
    });
  }
}
