import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnDestroy,
  OnInit,
  ViewChild,
} from '@angular/core';
import * as d3 from 'd3';
import { default as moment } from 'moment';
import { BsModalService } from 'ngx-bootstrap/modal';
import { Subscription } from 'rxjs';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { flyInAnimations } from '../../animations';
import { DashboardElementsOverlayComponent } from '../../dashboard-elements-overlay/dashboard-elements-overlay.modal';
import { DashboardElement } from '../../interfaces/dashboard-element';
import { HighlightService } from '../../services/highlight.service';
import {
  MediaOverviewType,
  OneUpApiService,
  TimespanType,
} from '../../services/one-up-api.service';
import { BarchartService } from '../../services/vis/barchart.service';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-hot-media',
  templateUrl: './hot-media.component.html',
  styleUrls: ['./hot-media.component.css', '../../app.component.css'],
  animations: flyInAnimations,
})
export class HotMediaComponent
  extends AbstractPage
  implements AfterViewInit, OnDestroy, OnInit, DashboardElement
{
  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();

  /**
   *  ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('hotmediachart', { static: false }) private chartContainer: ElementRef;

  /**
   *  define the charts view margin
   */
  margin = { top: 10, bottom: 55, left: 10, right: 10 };

  /**
   *  define the charts view width
   */
  width = 500 - this.margin.left - this.margin.right;

  /**
   *  define the charts view height
   */
  height = 220 - this.margin.top - this.margin.bottom;

  /** value that represents the active tab
   *    '0': show most popular media among all available topics
   *    '1': show most popular media among user selected topics (default)
   *    '2': show most popular media only for highlighted topic
   */
  activeTab = '1';

  /**
   *  holds the name of the highlighted topic
   */
  highlightedTopic = '';

  /**
   *  subscribes to the highlightedTopicChanged event to update highlightedTopic
   */
  highlightSubscription: Subscription;

  /**
   *  holds the color of the currently highlighted topic
   *  so that the bar chart can match that color
   */
  highlightedTopicColor = '';

  /**
   *  subscribes to changes in topic highlighting and updates
   *  highlightedTopicColor accordingly
   */
  highlightColorSubscription: Subscription;

  /**
   *  subscribes to clicking events on the bars
   *  to then show the most recent articles for the selected bar
   */
  changeMediumDetailSubscription: Subscription;

  /**
   *  indicates whether the view is displaying the bar chart (isDetailedVisible = false)
   *  or the list of most recent articles (isDetailedVisible = true)
   */
  isDetailedVisible = false;

  /**
   *  holds the mediums name once its corresponding bar has been clicked
   *  and the list view of the mediums most recent articles is fetched
   */
  detailMedium = '';

  /**
   *  holds an array of the most recent articles
   */
  mostRecentArticles = [];

  /**
   *  holds different values that are passed over for diagram creation
   */
  visOptions = {
    width: this.width,
    height: this.height,
    margin: this.margin,
    element: undefined,
    xColumnName: 'article_count',
    yColumnName: 'medium_name',
    scheme: this.colorService.barCharts.mediaOverview,
    xTitle: 'Veröffentlichungen',
    maxRange: undefined,
  };

  /**
   *  time period for which data is retrieved
   */
  selectedTimespan: TimespanType = TimespanType.totalPeriod;

  /**
   * inject the necessary services that are used in this component
   */
  constructor(
    private barchartService: BarchartService,
    private oneUpApiService: OneUpApiService,
    private highlightService: HighlightService,
    private modalService: BsModalService,
    private cdRef: ChangeDetectorRef
  ) {
    super();
  }

  /**
   *  On component initialization,
   *    - create listener, that refreshes the diagram when the topic selection changes
   */
  ngOnInit(): void {
    this.topicSelectionChanged.subscribe((value: any) => this.callAPIs());
  }

  /**
   *  Initializes the bar chart by
   *    - drawing the SVG
   *    - fetching the getHotMedia API
   *
   *  Subscribes to highlightSubscription, highlightColorSubscription, changeMediumDetailSubscription
   *  and defines the functionality when a change is received
   */
  ngAfterViewInit(): void {
    this.createSVG();

    this.oneUpApiService
      .getHotMedia(MediaOverviewType.userTags, this.selectedTimespan, this.user)
      .subscribe((response) => {
        this.visOptions.maxRange = response[0].article_count;
        this.barchartService.createUpdateHorizontalBars(this.visOptions, response);
      });

    this.highlightSubscription = this.highlightService.highlightedTopicChanged.subscribe(
      (value) => {
        this.activeTab = value ? '2' : '1';
        this.highlightedTopic = value ? value : '';
        this.callAPIs();
      }
    );

    this.highlightColorSubscription = this.highlightService.highlightedTopicColor.subscribe(
      (value) => {
        this.highlightedTopicColor = value;
      }
    );

    this.changeMediumDetailSubscription = this.barchartService.changeMediumDetailEmitter.subscribe(
      (value) => {
        if (value) {
          const obj = JSON.parse(value);

          if (!obj.activeOverlay) {
            this.changeInDetailview(obj.mediumName);
          }
        }
      }
    );
  }

  /**
   *  receives output from the timespan-selector component,
   *  sets the newly selected timespan and refreshes the data
   */
  timespanSelectionChanged(timespan) {
    this.selectedTimespan = TimespanType[timespan];
    this.callAPIs();
  }

  /**
   *  creates an SVG that will later be used as a base layer to display the chart
   */
  createSVG() {
    this.visOptions.element = this.barchartService.createSVGMedia(
      this.chartContainer.nativeElement,
      this.width,
      this.height,
      this.margin
    );
  }

  /**
   *  Depending on the activeTab and the displayed view (this.isDetailedVisible)
   *  this calls the API to either update the most popular media (getHotMedia)
   *  or the most recent articles (getRecentArticles)
   *  and creates (or updates) the corresponding view
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.visOptions.scheme = this.scheme;

    let mediaOverviewType;
    let query;
    switch (this.activeTab) {
      case '0':
        mediaOverviewType = MediaOverviewType.allTags;
        break;
      case '1':
        mediaOverviewType = MediaOverviewType.userTags;
        query = this.user;
        break;
      case '2':
        mediaOverviewType = MediaOverviewType.queryTag;
        query = this.highlightedTopic;
        break;
    }

    if (!this.isDetailedVisible) {
      this.oneUpApiService
        .getHotMedia(mediaOverviewType, this.selectedTimespan, query)
        .subscribe((response) => {
          this.visOptions.scheme =
            this.activeTab === '2'
              ? d3.scaleOrdinal().range([this.highlightedTopicColor])
              : this.colorService.barCharts.mediaOverview;
          this.visOptions.maxRange = response[0].article_count;
          this.barchartService.createUpdateHorizontalBars(this.visOptions, response);
          this.componentStillLoading = false;
          this.cdRef.detectChanges();
        });
    } else {
      this.oneUpApiService
        .getRecentArticles(mediaOverviewType, 5, query, this.detailMedium)
        .subscribe((response) => {
          const data = JSON.parse(JSON.stringify(response));
          for (const article of data) {
            article.date_of_publication = moment(article.date_of_publication).format(
              'DD. MMMM YYYY'
            );
          }

          this.mostRecentArticles = data;
          this.componentStillLoading = false;
          this.cdRef.detectChanges();
        });
    }
  }

  /**
   *  Changes the toggle mode and calls the API to get the corresponding data.
   */
  changeSelection(selectedTab) {
    this.activeTab = selectedTab;
    this.callAPIs();
  }

  /**
   *  Unsubscribes from active subscriptions once the component is destroyed
   */
  ngOnDestroy() {
    this.highlightSubscription.unsubscribe();
    this.highlightColorSubscription.unsubscribe();
    this.changeMediumDetailSubscription.unsubscribe();
  }

  /**
   *  Toggles the view between popular media and recent articles
   *  and re-initializes the graph/view
   */
  changeInDetailview(name: string) {
    this.isDetailedVisible = !this.isDetailedVisible;
    this.detailMedium = name;

    this.cdRef.detectChanges();

    if (!this.isDetailedVisible) {
      this.createSVG();
    }

    this.callAPIs();
  }

  /**
   *  Capitalizes the first letter of a mediums name (unless it is a website)
   */
  capitalize(str: string): string {
    return str
      .split(' ')
      .map((word) => {
        if (word.indexOf('.de') < 0 && word.indexOf('.com') < 0) {
          if (word[0] === '(') {
            return '(' + word[1].toUpperCase() + word.substring(2);
          } else {
            return word[0].toUpperCase() + word.substring(1);
          }
        } else {
          return word;
        }
      })
      .join(' ');
  }

  /**
   *  Logs activity to the backend when the user opens an article
   */
  logUserActivity(title: string) {
    this.oneUpApiService
      .postUserActivity(this.user, 'userOpenedArticle', {
        activity: 'userOpenedArticle',
        articleName: title,
      })
      .subscribe(() => {});
  }

  /**
   *  Passes data to the modalService and opens diagram in the DashboardElementsOverlayComponent
   */
  openOverlay() {
    const initialState = {
      title: 'Medien-Übersicht',
      wikiName: 'Medienübersicht',
      visOptions: JSON.parse(JSON.stringify(this.visOptions)),
      user: this.user,
      highlightedTopic: this.highlightedTopic,
      highlightedTopicColor: this.highlightedTopicColor,
      activeTab: this.activeTab,
      topicSelectionChanged: this.topicSelectionChanged,
      selectedTimespan: this.selectedTimespan,
    };
    this.modalService.show(DashboardElementsOverlayComponent, { initialState });
  }
}
