import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnInit,
  ViewChild,
} from '@angular/core';
import { BsModalService } from 'ngx-bootstrap/modal';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { DashboardElementsOverlayComponent } from '../../dashboard-elements-overlay/dashboard-elements-overlay.modal';
import { DashboardElement } from '../../interfaces/dashboard-element';
import { MediaOverviewType, OneUpApiService } from '../../services/one-up-api.service';
import { BarchartService } from '../../services/vis/barchart.service';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-hot-topics',
  templateUrl: './hot-topics.component.html',
  styleUrls: ['./hot-topics.component.css', '../../app.component.css'],
})
export class HotTopicsComponent
  extends AbstractPage
  implements AfterViewInit, OnInit, DashboardElement
{
  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();

  /**
   *  ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('hottopicbar') private chartContainer: ElementRef;

  /**
   *  define the charts view margin
   */
  margin = { top: 10, bottom: 65, left: 20, right: 20 };

  /**
   *  define the charts view height
   */
  height = 240 - this.margin.top - this.margin.bottom;

  /**
   *  define the charts view width
   */
  width = 500 - this.margin.left - this.margin.right;

  /** value that represents the active tab
   *    '0': number of weekly changes in articles reporting on the selected topics (default)
   *    '1': number of monthly changes in articles reporting on the selected topics
   */
  activeTab = '0';

  /**
   *  holds different values that are passed over for diagram creation
   */
  visOptions = {
    width: this.width,
    height: this.height,
    margin: this.margin,
    xColumnLeftName: 'diff_last_week',
    xColumnRightName: 'diff_last_week',
    yColumnName: 'tag_name',
    element: undefined,
    scheme: this.colorService.barCharts.blue,
    maxRange: 1,
    xTitle: 'Veröffentlichungen – Differenz zur Vorwoche',
    xTitleLeft: '',
    xTitleRight: '',
    negativeScale: true,
  };

  /**
   *  Holds the latest response from getHotTopics API
   *  so we can toggle between weekly/monthly view without fetching the API again
   */
  private latestResponse;

  /**
   *  Holds the latest response from getUserTopics API
   *  so we can toggle between weekly/monthly view without fetching the API again
   */
  private latestUserFavoriteTopics;

  /**
   *  inject the necessary services that are used in this component
   */
  constructor(
    private barchartService: BarchartService,
    private oneUpApiService: OneUpApiService,
    private modalService: BsModalService,
    private cdRef: ChangeDetectorRef
  ) {
    super();
  }

  /**
   *  On component initialization, subscribe to the topicSelectionChanged listener to refresh data
   */
  ngOnInit(): void {
    this.topicSelectionChanged.subscribe((value: any) => this.callAPIs());
  }

  /**
   *  Initializes the bar chart by
   *    - drawing the SVG
   *    - refreshing the API data
   */
  ngAfterViewInit(): void {
    this.visOptions.element = this.barchartService.createSVGMedia(
      this.chartContainer.nativeElement,
      this.width,
      this.height,
      this.margin
    );
    this.callAPIs();
  }

  /**
   *  Changes the toggle mode, updates chart labels and updates the chart
   */
  changeSelection(selectedTab) {
    this.visOptions.scheme = this.scheme;
    this.activeTab = selectedTab;
    if (this.activeTab === '0') {
      this.visOptions.xColumnLeftName = 'diff_last_week';
      this.visOptions.xColumnRightName = 'diff_last_week';
      this.visOptions.xTitle = 'Veröffentlichungen – Differenz zur Vorwoche';
    } else {
      this.visOptions.xColumnLeftName = 'diff_last_month';
      this.visOptions.xColumnRightName = 'diff_last_month';
      this.visOptions.xTitle = 'Veröffentlichungen – Differenz zum Vormonat';
    }
    this.barchartService.createUpdatePyramidBars(
      this.visOptions,
      this.latestResponse,
      this.latestUserFavoriteTopics
    );
  }

  /**
   *  Fetches the getHotTopics API and draws the pyramid chart
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.visOptions.scheme = this.scheme;
    this.oneUpApiService
      .getHotTopics(this.user, MediaOverviewType.userTags)
      .subscribe((response) => {
        this.latestResponse = response;
        this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
          this.latestUserFavoriteTopics = userFavoriteTopics;
          this.barchartService.createUpdatePyramidBars(
            this.visOptions,
            this.latestResponse,
            userFavoriteTopics
          );
          this.componentStillLoading = false;
          this.cdRef.detectChanges();
        });
      });
  }

  /**
   *  Passes data to the modalService and opens diagram in the DashboardElementsOverlayComponent
   */
  openOverlay() {
    const initialState = {
      title: 'Angesagte Themen im Vergleich',
      wikiName: 'Angesagte-Themen',
      visOptions: JSON.parse(JSON.stringify(this.visOptions)),
      user: this.user,
      activeTab: this.activeTab,
      topicSelectionChanged: this.topicSelectionChanged,
    };
    this.modalService.show(DashboardElementsOverlayComponent, { initialState });
  }
}
