import { ChangeDetectorRef, Component, OnInit } from '@angular/core';
import { MatSelectChange } from '@angular/material/select';
import { default as moment } from 'moment';
import { Meals } from 'src/app/interfaces/mensa';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { DashboardElement } from '../../interfaces/dashboard-element';
import { MensaService } from '../../services/mensa.service';

/**
 * This component loads the menu of the mensa at the University of Potsdam.
 */
@Component({
  selector: 'app-mensa',
  templateUrl: './mensa.component.html',
  styleUrls: ['./mensa.component.css', '../../app.component.css'],
})
export class MensaComponent extends AbstractPage implements OnInit, DashboardElement {
  /**
   * Initial location
   */
  location = 'Golm';
  /**
   * processed meals
   */
  sortedMeals = Object.create(Object.prototype);
  /**
   * The unprocessed meals returned by the API.
   */
  private meal: Meals;

  /**
   * Constructor for the component which is used to initialize the corresponding service.
   *
   * @param service to communicate via service with the API
   */
  constructor(private service: MensaService, private cdRef: ChangeDetectorRef) {
    super();
  }

  /**
   * During initialization the menu is retrieved from the UP-API.
   */
  ngOnInit(): void {
    this.callAPIs();
  }

  /**
   * Calls the service to interact with the API.
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.service.getMeal(this.location).subscribe((meal) => {
      this.processMeal(meal);
      this.componentStillLoading = false;
      this.cdRef.detectChanges();
    });
  }

  /**
   * Processes the incoming data from the API to display it in the view.
   *
   * @param meal The unfiltered answer from the API
   */
  processMeal(meal: Meals) {
    meal.meal.sort((a, b) => a.order - b.order);
    this.meal = meal;
    this.constructSortedMeals();
  }

  /**
   * Callable function to change the selected destination. This creates a new call to the API.
   *
   * @param event emitted event which includes the selected location
   */
  changeSelection(event: MatSelectChange) {
    this.location = event.value;
    this.callAPIs();
  }

  /**
   * Filter the menu to get a displayable list with title and description.
   */
  constructSortedMeals() {
    this.sortedMeals = {};
    const filteredMeals = this.meal.meal.filter(
      (m) =>
        m.title !== 'Info' &&
        m.title !== 'Information' &&
        m.description !== 'a' &&
        m.description !== '.'
    );
    filteredMeals.forEach((fMeal) => {
      if (moment(fMeal.date).format('MM DD YYYY') === moment().format('MM DD YYYY')) {
        if (!this.sortedMeals.hasOwnProperty(fMeal.title)) {
          this.sortedMeals[fMeal.title] = [];
        }
        this.sortedMeals[fMeal.title].push(fMeal.description);
      }
    });
  }

  /**
   * Extracts the keys for the user interface
   */
  getValidMealsKeys() {
    return Object.keys(this.sortedMeals);
  }

  /**
   * Gets a description for the key
   *
   * @param key to the meal
   */
  getValidMealsValue(key: string) {
    return this.sortedMeals[key];
  }
}
