import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnInit,
  ViewChild,
} from '@angular/core';
import { BsModalService } from 'ngx-bootstrap/modal';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { DashboardElementsOverlayComponent } from '../../dashboard-elements-overlay/dashboard-elements-overlay.modal';
import { DashboardElement } from '../../interfaces/dashboard-element';
import {
  MediaOverviewType,
  OneUpApiService,
  TimespanType,
} from '../../services/one-up-api.service';
import { BarchartService } from '../../services/vis/barchart.service';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-regional-influence',
  templateUrl: './regional-influence.component.html',
  styleUrls: ['./regional-influence.component.css', '../../app.component.css'],
})
export class RegionalInfluenceComponent
  extends AbstractPage
  implements AfterViewInit, OnInit, DashboardElement
{
  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();

  /**
   *  ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('pyramidchart') private chartContainer: ElementRef;

  /**
   *  define the charts view margin
   */
  margin = { top: 10, bottom: 50, left: 10, right: 10 };

  /**
   *  define the charts view width
   */
  width = 500 - this.margin.left - this.margin.right;

  /**
   *  define the charts view height
   */
  height = 220 - this.margin.top - this.margin.bottom;

  /**
   *  holds different values that are passed over for diagram creation
   */
  visOptions = {
    width: this.width,
    height: this.height,
    margin: this.margin,
    element: undefined,
    yColumnName: 'tag_name',
    xColumnLeftName: 'regional_count',
    xColumnRightName: 'national_count',
    scheme: this.colorService.barCharts.blue,
    xTitle: 'Veröffentlichungen',
    xTitleLeft: 'Regional',
    xTitleRight: 'Überregional',
  };

  /**
   *  time period for which data is retrieved
   */
  selectedTimespan: TimespanType = TimespanType.totalPeriod;

  /**
   *  inject the necessary services that are used in this component
   */
  constructor(
    private barchartService: BarchartService,
    private oneUpApiService: OneUpApiService,
    private modalService: BsModalService,
    private cdRef: ChangeDetectorRef
  ) {
    super();
  }

  /**
   *  On component initialization, subscribe to the topicSelectionChanged listener to refresh data
   */
  ngOnInit(): void {
    this.topicSelectionChanged.subscribe((value: any) => this.callAPIs());
  }

  /**
   *  Initializes the pyramid chart by
   *    - drawing the SVG
   *    - refreshing the API data
   */
  ngAfterViewInit(): void {
    this.visOptions.element = this.barchartService.createSVGMedia(
      this.chartContainer.nativeElement,
      this.width,
      this.height,
      this.margin
    );
    this.callAPIs();
  }

  /**
   *  receives output from the timespan-selector component,
   *  sets the newly selected timespan and refreshes the data
   */
  timespanSelectionChanged(timespan) {
    this.selectedTimespan = TimespanType[timespan];
    this.callAPIs();
  }

  /**
   *  Fetches the getUserTagsWithArticleCount API and draws (or updates) the pyramid chart
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.visOptions.scheme = this.scheme;

    this.oneUpApiService
      .getUserTagsWithArticleCount(this.user, MediaOverviewType.userTags, this.selectedTimespan)
      .subscribe((response) => {
        this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
          this.barchartService.createUpdatePyramidBars(
            this.visOptions,
            response,
            userFavoriteTopics
          );
          this.componentStillLoading = false;
          this.cdRef.detectChanges();
        });
      });
  }

  /**
   *  Passes data to the modalService and opens diagram in the DashboardElementsOverlayComponent
   */
  openOverlay() {
    const initialState = {
      title: 'Regionale Themen',
      wikiName: 'Regionale-Themen',
      visOptions: JSON.parse(JSON.stringify(this.visOptions)),
      user: this.user,
      topicSelectionChanged: this.topicSelectionChanged,
      selectedTimespan: this.selectedTimespan,
    };
    this.modalService.show(DashboardElementsOverlayComponent, { initialState });
  }
}
