import { Component, HostListener, Inject, OnInit } from '@angular/core';
import { FormControl, Validators } from '@angular/forms';
import { MatDialog, MAT_DIALOG_DATA } from '@angular/material/dialog';
import { MatTableDataSource } from '@angular/material/table';
import { convertToClassName } from '../../../util';

/**
 * Data representation of a Tip for Tipp des Tages
 */
class Tip {
  /**
   * ID for db purpose
   */
  id: number;
  /**
   * Text of the tip
   */
  content: string;
  /**
   * ID of the faculty
   */
  facId: Faculty;
  /**
   * Begin of the time duration in which this will be shown
   */
  since: Date;
  /**
   * How long the tip will be shown.
   */
  duration: number;
  /**
   * How many times the tip is shown.
   */
  shown: number;

  /**
   * Constructor of the data representation
   *
   * @param id ID for db purpose
   * @param content Text of the tip
   * @param facId ID of the faculty
   * @param since Begin of the time duration in which this will be shown.
   * @param duration How long the tip will be shown.
   * @param shown How many times the tip is shown.
   */
  constructor(
    id: number,
    content: string,
    facId: Faculty,
    since: Date,
    duration: number,
    shown: number
  ) {
    this.id = id;
    this.content = content;
    this.facId = facId;
    this.since = since;
    this.duration = duration;
    this.shown = shown;
  }
}

/**
 * Controller of the form to enter a tip.
 */
interface TipFormControl {
  /**
   * Form control to insert the text of the tip.
   */
  contentFormControl: FormControl;
  /**
   * Form control to insert the faculty of the tip.
   */
  facIDFormControl: FormControl;
  /**
   * Form control to insert when the tip is shown.
   */
  sinceFormControl: FormControl;
  /**
   * Form control to insert how long the tip is shown.
   */
  durationFormControl: FormControl;
}

/**
 * Data representation of a faculty which is connected to a tip.
 */
interface Faculty {
  /**
   * ID representation in the database.
   */
  realVal: number;
  /**
   * Name of the faculty.
   */
  shownVal: string;
}

/**
 * Data representation of the deleted data.
 */
interface DelDialogData {
  /**
   * Array with all tips to be deleted.
   */
  tips: Tip[];
  /**
   * Selected tips will be deleted. This field indicates whether a tip is selected.
   */
  selects: boolean[];
}

/**
 * Data representation of possible errors in the form.
 */
interface ErrorState {
  /**
   * Error: No text inserted.
   */
  cntErr: boolean;
  /**
   * Error: No faculty inserted.
   */
  facErr: boolean;
  /**
   * Error: No start date inserted.
   */
  sncErr: boolean;
}

/**
 * static column color
 */
const editCol = 'blue';

/**
 * The form to insert a Tipp des Tages
 */
@Component({
  selector: 'app-tdt-form',
  templateUrl: './tdt-form.component.html',
  styleUrls: ['./tdt-form.component.css', '../../../app.component.css'],
})
export class TdtFormComponent implements OnInit {
  dc: string[] = ['select', 'content', 'facId', 'since', 'duration', 'shown'];
  faculties: Faculty[] = [
    { realVal: 0, shownVal: 'facId0' },
    { realVal: 1, shownVal: 'facId1' },
    { realVal: 2, shownVal: 'facId2' },
    { realVal: 3, shownVal: 'facId3' },
    { realVal: 4, shownVal: 'facId4' },
  ];
  allTips: Tip[] = [
    new Tip(0, 'Tip 1', this.faculties[0], new Date('1998-10-22'), 200, 0),
    new Tip(1, 'Tip 2', this.faculties[1], new Date('1998-10-23'), 100, 0),
    new Tip(2, 'Tip 3', this.faculties[2], new Date('1998-10-24'), 0, 0),
    new Tip(3, 'Tip 4', this.faculties[3], new Date('1998-10-25'), 150, 0),
  ];
  orgTips: Tip[] = [
    new Tip(0, 'Tip 1', this.faculties[0], new Date('1998-10-22'), 200, 0),
    new Tip(1, 'Tip 2', this.faculties[1], new Date('1998-10-23'), 100, 0),
    new Tip(2, 'Tip 3', this.faculties[2], new Date('1998-10-24'), 0, 0),
    new Tip(3, 'Tip 4', this.faculties[3], new Date('1998-10-25'), 150, 0),
  ];
  editCross = [-1, -1, 0];
  newFormControl = this.createTipFormControl();
  matds = new MatTableDataSource<Tip>();
  newTip = this.createEmptyTip();
  selmaster = false;
  selarr: boolean[] = [false, false, false, false];
  tmpTips: Tip[] = [];
  disdel = true;
  disadur: boolean[] = [false]; // ??start value nessessary
  tadd = false;
  tchange = false;
  pickerOpen = 0;
  errors: ErrorState[] = [];

  constructor(public dialog: MatDialog) {}

  @HostListener('document:mousedown')
  checkCrossReset(): void {
    if (!(this.editCross[2] + this.pickerOpen) && this.validCross()) {
      if (this.editCross[0] !== -1) {
        if (!this.disadur[0] || this.allTips[this.editCross[0]].duration === null)
          this.allTips[this.editCross[0]].duration = 0; // !!code duplication
        this.checkForChanges(this.allTips[this.editCross[0]]); // ??reduce to just checking one cell
      }
      this.editCross = [-1, -1, 0];
    } else this.editCross[2] = 0;
  }

  ngOnInit(): void {
    /*fetch('http://localhost:4321/tdt/stats')
      .then(res=>res.json())
      .then(res=>res.map((res:any,n)=>this.loadAPI(res,n++)));*/
    this.matds = new MatTableDataSource<Tip>(this.allTips);
  }

  loadAPI(tip: any, n: number): void {
    this.allTips.push(
      new Tip(
        n,
        tip.content,
        this.faculties[tip.facId],
        new Date(tip.since[0] + '-' + tip.since[1] + '-' + tip.since[2]),
        tip.duration,
        tip.shown
      )
    ); // ??date im string def. lassen
    this.orgTips.push(this.getTipValue(n)); // !!shit code
    this.selarr.push(false);
  }

  changeEditCross(i: number, j: number): void {
    if (this.editCross[0] === -1 || this.validCross()) {
      // !!expensive recurring
      if (!this.tchange && !(this.editCross[0] === i && this.editCross[1] === j)) {
        this.tmpTips[0] = this.getTipValue(i); // §§fix for reset .rem=?
        if (this.editCross[0] !== -1) {
          if (!this.disadur[0] || this.allTips[this.editCross[0]].duration === null)
            this.allTips[this.editCross[0]].duration = 0; // !!ugly and broken
        }
        this.disadur[0] = this.allTips[i].duration !== 0; // §§fix logic
      }
      if (this.editCross[0] !== -1) this.checkForChanges(this.allTips[this.editCross[0]]);
      this.editCross = [i, j, +!this.tchange];
    }
  }

  checkForChanges(tip: Tip): void {
    if (tip.id !== -1) {
      const ttip = this.orgTips[tip.id]; // !!find better way of doing this .ref_strat=reset
      if (tip.content !== ttip.content) this.colourChanges('cnt' + tip.id, editCol);
      else this.colourChanges('cnt' + tip.id, 'black'); // !!ugly
      if (tip.facId.realVal !== ttip.facId.realVal) this.colourChanges('fac' + tip.id, editCol);
      else this.colourChanges('fac' + tip.id, 'black');
      // ??way too verbose
      if (
        tip.since.getDate() !== ttip.since.getDate() ||
        tip.since.getMonth() !== ttip.since.getMonth() ||
        tip.since.getFullYear() !== ttip.since.getFullYear()
      )
        this.colourChanges('snc' + tip.id, editCol);
      else this.colourChanges('snc' + tip.id, 'black');
      if (tip.duration !== ttip.duration) this.colourChanges('dur' + tip.id, editCol);
      else this.colourChanges('dur' + tip.id, 'black');
    }
  }

  colourChanges(dest: string, col: string): void {
    const cls = document.getElementsByClassName(dest);
    cls[0].setAttribute('style', 'color:' + col);
  }

  selectAll(): void {
    for (let i = 0; i < this.selarr.length; i++) this.selarr[i] = this.selmaster;
    if (this.selarr.length !== 0) this.disdel = !this.selmaster;
  }

  checkAllSelected(): void {
    this.selmaster = true;
    this.disdel = true;
    if (this.selarr.length === 0) {
      this.selmaster = false;
      return;
    }
    for (const i of this.selarr) {
      if (i === false) this.selmaster = false;
      else this.disdel = false;
    }
  }

  openTipAdd(): void {
    this.errors = [];
    this.errors.push({ cntErr: false, facErr: false, sncErr: false });
    this.tadd = true;
  }

  addNewTip(): void {
    if (!this.disadur[0] || this.newTip.duration == null) this.newTip.duration = 0;
    this.markAllFormTouched();
    const cntInvalid =
      !this.laotzu(this.newTip.content) && !this.tipAlreadyExists(-1, this.newTip.content);
    if (
      cntInvalid &&
      this.newTip.since !== null &&
      this.newTip.facId !== null &&
      this.newTip.duration >= 0
    ) {
      this.allTips.push({
        id: -1,
        content: this.newTip.content,
        facId: this.newTip.facId,
        since: this.newTip.since,
        duration: this.newTip.duration,
        shown: this.newTip.shown,
      });
      this.matds = new MatTableDataSource<Tip>(this.allTips);
      this.selarr.push(false);
      this.selmaster = false;
      this.closeTipAdd();
    } else {
      this.errors[0].cntErr = !cntInvalid;
      this.errors[0].facErr = this.newTip.facId === null;
      this.errors[0].sncErr = this.newTip.since === null;
    }
  }

  changeTips(): void {
    this.tchange = true;
    this.tadd = false;
    this.tmpTips = [];
    this.disadur = [];
    this.errors = [];
    for (const tip of this.allTips) {
      // ??copy only if selected
      this.tmpTips.push({
        id: tip.id,
        content: tip.content,
        facId: tip.facId,
        since: tip.since,
        duration: tip.duration,
        shown: tip.shown,
      }); // !!value
      this.disadur.push(tip.duration !== 0);
      this.errors.push({ cntErr: false, facErr: false, sncErr: false });
    }
  }

  closeTipAddDialog(): void {
    const dialogRef = this.dialog.open(TdtFormConfComponent, {
      width: '25%',
      height: '20%',
      data: 'Schließen',
    });
    dialogRef.afterClosed().subscribe((res) => {
      if (res) this.closeTipAdd();
    });
  }

  closeTipAdd(): void {
    this.markAllFormUntouched();
    this.newTip = this.createEmptyTip();
    this.disadur[0] = false;
    this.tadd = false;
  }

  closeTipChange(): void {
    this.tchange = false;
    let i = 0;
    for (const tip of this.allTips) {
      if (!this.disadur[i] || tip.duration === null) this.allTips[i].duration = 0;
      const cntInvalid = this.laotzu(tip.content) || this.tipAlreadyExists(i, tip.content);
      if (cntInvalid || tip.facId == null || tip.since == null) {
        // !!reduce this shenanigans
        this.errors[i].cntErr = cntInvalid;
        this.errors[i].facErr = tip.facId === null;
        this.errors[i].sncErr = tip.since === null;
        this.tchange = true;
      } else this.checkForChanges(tip);
      i++;
    }
  }

  resetTipChange(): void {
    // ??copy only if selected
    const dialogRef = this.dialog.open(TdtFormConfComponent, {
      width: '25%',
      height: '20%',
      data: 'Zurücksetzen',
    });
    dialogRef.afterClosed().subscribe((res) => {
      if (res) {
        this.allTips = this.tmpTips;
        for (let i = 0; i < this.allTips.length; i++)
          this.disadur[i] = this.allTips[i].duration !== 0;
        this.matds = new MatTableDataSource<Tip>(this.allTips);
      }
    });
  }

  resetCross(): void {
    this.pickerOpen = 1;
    const dialogRef = this.dialog.open(TdtFormConfComponent, {
      width: '25%',
      height: '20%',
      data: 'Zurücksetzen',
    });
    dialogRef.afterClosed().subscribe((res) => {
      // !!reduce
      if (res) {
        this.allTips[this.editCross[0]] = {
          id: this.tmpTips[0].id,
          content: this.tmpTips[0].content,
          facId: this.tmpTips[0].facId,
          since: this.tmpTips[0].since,
          duration: this.tmpTips[0].duration,
          shown: this.tmpTips[0].shown,
        }; // §§fix for reset .rem=?
        this.disadur[0] = this.tmpTips[0].duration !== 0;
        this.matds = new MatTableDataSource<Tip>(this.allTips); // ??reduce recurring pattern
      }
      this.pickerOpen = 0;
    });
  }

  openDelDialog(): void {
    const dialogRef = this.dialog.open(TdtFormDialogComponent, {
      width: '50%',
      height: '50%',
      data: { tips: this.allTips, selects: this.selarr },
    });
    dialogRef.afterClosed().subscribe((res) => {
      this.allTips = res.tips;
      this.selarr = res.selects;
      this.checkAllSelected();
      this.matds = new MatTableDataSource<Tip>(this.allTips);
    });
  }

  markAllFormTouched(): void {
    this.newFormControl.contentFormControl.markAsTouched();
    this.newFormControl.facIDFormControl.markAsTouched();
    this.newFormControl.sinceFormControl.markAsTouched();
    this.newFormControl.durationFormControl.markAsTouched();
  }

  markAllFormUntouched(): void {
    this.newFormControl.contentFormControl.markAsUntouched();
    this.newFormControl.facIDFormControl.markAsUntouched();
    this.newFormControl.sinceFormControl.markAsUntouched();
    this.newFormControl.durationFormControl.markAsUntouched();
  }

  createFormControl(): FormControl {
    return new FormControl('', [Validators.required]);
  }

  createTipFormControl(): TipFormControl {
    return {
      contentFormControl: this.createFormControl(),
      facIDFormControl: this.createFormControl(),
      sinceFormControl: this.createFormControl(),
      durationFormControl: this.createFormControl(),
    }; // !!kill last form
  }

  createEmptyTip(): Tip {
    return {
      id: -1,
      content: '',
      facId: null,
      since: new Date(),
      duration: null,
      shown: 0,
    };
  }

  getTipValue(n: number): Tip {
    // !!NO !!BAD
    return {
      id: this.allTips[n].id,
      content: this.allTips[n].content,
      facId: this.allTips[n].facId,
      since: this.allTips[n].since,
      duration: this.allTips[n].duration,
      shown: this.allTips[n].shown,
    };
  }

  getDateFormat(date: Date): string {
    if (date !== null)
      return date.getDate() + '. ' + (date.getMonth() + 1) + '. ' + date.getFullYear();
    else return 'invalid date';
  }

  laotzu(x: string): boolean {
    return !convertToClassName(x).length;
  } // ??why not vanilla .unused
  tipAlreadyExists(j: number, x: string): boolean {
    let i = 0;
    for (const tip of this.allTips) {
      // !!RUNTIME!!
      if (
        convertToClassName(tip.content).toLowerCase() === convertToClassName(x).toLowerCase() &&
        j !== i
      )
        return true;
      i++;
    }
    return false;
  }

  validCross(): boolean {
    if (this.editCross[0] === -1) return true;
    return (
      !this.laotzu(this.allTips[this.editCross[0]].content) &&
      this.allTips[this.editCross[0]].since !== null &&
      !this.tipAlreadyExists(this.editCross[0], this.allTips[this.editCross[0]].content)
    );
  } // !!add everything else
  sendUpdates(): void {
    const dialogRef = this.dialog.open(TdtFormSendComponent, {
      width: '75%',
      height: '75%',
      data: [this.allTips, this.orgTips],
    });
    dialogRef.afterClosed().subscribe(/*res => {}*/);
  }
}

/**
 * Dialog to delete data.
 */
@Component({
  selector: 'app-tdt-form-dialog',
  templateUrl: './tdt-form-dialog.component.html',
})
export class TdtFormDialogComponent {
  ntips: Tip[] = [];
  dtips: Tip[] = [];
  dds = new MatTableDataSource<Tip>();
  nselects: boolean[] = [];
  dc: string[] = ['content', 'facId', 'since', 'duration', 'shown']; // !!das ist kein code sondern müll
  constructor(@Inject(MAT_DIALOG_DATA) public data: DelDialogData) {
    for (let i = 0; i < data.selects.length; i++) {
      if (!this.data.selects[i]) {
        this.ntips.push(data.tips[i]);
        this.nselects.push(data.selects[i]);
      } else this.dtips.push(data.tips[i]);
    }
    this.dds = new MatTableDataSource<Tip>(this.dtips);
  }

  deleteSelects(): void {
    this.data.tips = this.ntips;
    this.data.selects = this.nselects;
  }
}

/**
 * Config the Tip des Tages Form with data.
 */
@Component({
  selector: 'app-tdt-form-conf',
  templateUrl: './tdt-form-conf.component.html',
})
export class TdtFormConfComponent {
  constructor(@Inject(MAT_DIALOG_DATA) public data: string) {}
}

/**
 * Conformation dialog to insert Tipp des Tages.
 */
@Component({
  selector: 'app-tdt-form-send',
  templateUrl: './tdt-form-send.component.html',
})
export class TdtFormSendComponent {
  dc: string[] = ['content', 'facId', 'since', 'duration', 'shown']; // !!recurring. pls stop
  ntips: Tip[] = [];
  ctips: Tip[] = [];
  rtips: Tip[] = [];
  addds = new MatTableDataSource<Tip>();
  chds = new MatTableDataSource<Tip>();
  rmds = new MatTableDataSource<Tip>();

  constructor(@Inject(MAT_DIALOG_DATA) public data: [Tip[], Tip[]]) {
    // ??maybe [Tip[],number[]]
    let i = 0;
    for (const tt of data[0]) {
      // !!find inaccuracies
      while (i < tt.id) {
        this.rtips.push(data[1][i]);
        i++;
      }
      if (tt.id === -1) this.ntips.push(tt);
      else {
        if (this.isChanged(tt, data[1][tt.id])) this.ctips.push(tt);
      }
      i++;
    }
    this.addds = new MatTableDataSource<Tip>(this.ntips);
    this.chds = new MatTableDataSource<Tip>(this.ctips);
    this.rmds = new MatTableDataSource<Tip>(this.rtips);
  }

  isChanged(t1: Tip, t2: Tip): boolean {
    // ??warum so !!na ganz einfach. das brauch ich dann fürs mergehandling
    return (
      t1.content !== t2.content ||
      t1.facId.realVal !== t2.facId.realVal ||
      t1.since.getDate() !== t2.since.getDate() ||
      t1.since.getMonth() !== t2.since.getMonth() ||
      t1.since.getFullYear() !== t2.since.getFullYear() ||
      t1.duration !== t2.duration
    );
  }
}
