import {
  AfterViewInit,
  ChangeDetectorRef,
  Component,
  ElementRef,
  EventEmitter,
  Input,
  OnDestroy,
  OnInit,
  ViewChild,
} from '@angular/core';
import { BsModalService } from 'ngx-bootstrap/modal';
import { Subscription } from 'rxjs';
import { AbstractPage } from '../../abstract-page/abstract-page';
import { DashboardElementsOverlayComponent } from '../../dashboard-elements-overlay/dashboard-elements-overlay.modal';
import { HighlightService } from '../../services/highlight.service';
import { MediaOverviewType, OneUpApiService } from '../../services/one-up-api.service';
import { LineGraphService } from '../../services/vis/line-graph.service';

/**
 * defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-topic-reach',
  templateUrl: './topic-reach.component.html',
  styleUrls: ['./topic-reach.component.css', '../../app.component.css'],
})
export class TopicReachComponent extends AbstractPage implements AfterViewInit, OnDestroy, OnInit {
  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Input() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();

  /**
   * ElementRef to the DOM element in which the SVG + diagram are displayed
   */
  @ViewChild('linegraph') private chartContainer: ElementRef;

  /**
   * value that represents the active tab
   *   '0': show article count of all available topics
   *   '1': show article count of user selected topics (default)
   *   '2': show article count of highlighted topic
   */
  activeTab = '1';

  /**
   * holds the name of the highlighted topic
   */
  highlightedTopic = '';

  /**
   * subscribes to the highlightedTopicChanged event to update highlightedTopic
   */
  highlightSubscription: Subscription;

  /**
   * define the charts view margin
   */
  margin = { top: 10, bottom: 20, left: 40, right: 25 };

  /**
   * define the charts view width
   */
  width = 500 - this.margin.left - this.margin.right;

  /**
   * define the charts view height
   */
  height = 200 - this.margin.top - this.margin.bottom;

  /**
   * holds different values that are passed over for diagram creation
   */
  visOptions = {
    width: this.width,
    height: this.height,
    margin: this.margin,
    element: undefined,
    yColumnName: 'reach',
    xColumnName: 'date',
    scheme: this.colorService.barCharts.blue,
  };

  /**
   * inject the necessary services that are used in this component
   */
  constructor(
    private oneUpApiService: OneUpApiService,
    private lineGraphService: LineGraphService,
    private highlightService: HighlightService,
    private modalService: BsModalService,
    private cdRef: ChangeDetectorRef
  ) {
    super();
  }

  /**
   * On component initialization, create listener
   *  - that refreshes the diagram when a topic is highlighted
   *  - that refreshes the diagram when the topic selection changes
   */
  ngOnInit(): void {
    this.highlightSubscription = this.highlightService.highlightedTopicChanged.subscribe(
      (value) => {
        this.activeTab = value ? '2' : '1';
        this.highlightedTopic = value ? value : '';
        this.callAPIs();
      }
    );

    this.topicSelectionChanged.subscribe((value: any) => this.callAPIs());
  }

  /**
   * Once the component is initialized, create an SVG that will later display the diagram.
   * Then, call the APIs to initialize and create the diagram.
   */
  ngAfterViewInit() {
    this.visOptions.element = this.lineGraphService.createSVG(
      this.chartContainer.nativeElement,
      this.visOptions
    );
    this.callAPIs();
  }

  /**
   * Unsubscribes from active subscriptions once the component is destroyed
   */
  ngOnDestroy() {
    this.highlightSubscription.unsubscribe();
  }

  /**
   * Changes the toggle mode and calls the API to get the corresponding data.
   *
   * @param event – holds the value of the selected mat-button-toggle
   */
  changeSelection(selectedTab) {
    this.activeTab = selectedTab;
    this.callAPIs();
  }

  /**
   * Depending on the activeTab, this functions build and calls the getTagsTimeview API
   * and creates (or updates) the article count line graph
   */
  callAPIs() {
    this.componentStillLoading = true;
    this.visOptions.scheme = this.scheme;

    let mediaOverviewType;
    let query;
    switch (this.activeTab) {
      case '0':
        mediaOverviewType = MediaOverviewType.allTags;
        break;
      case '1':
        mediaOverviewType = MediaOverviewType.userTags;
        query = this.user;
        break;
      case '2':
        mediaOverviewType = MediaOverviewType.queryTag;
        query = this.highlightedTopic;
        break;
    }

    this.oneUpApiService.getTagsTimeview(mediaOverviewType, query).subscribe((response) => {
      this.oneUpApiService.getUserTopics(this.user).subscribe((userFavoriteTopics) => {
        this.lineGraphService.createUpdateLineGraph(this.visOptions, response, userFavoriteTopics);
        this.componentStillLoading = false;
        this.cdRef.detectChanges();
      });
    });
  }

  /**
   * Passes data to the modalService and opens diagram in the DashboardElementsOverlayComponent
   */
  openOverlay() {
    const initialState = {
      title: 'Themenreichweite',
      wikiName: 'Reichweite-pro-Thema',
      visOptions: JSON.parse(JSON.stringify(this.visOptions)),
      user: this.user,
      highlightedTopic: this.highlightedTopic,
      activeTab: this.activeTab,
      topicSelectionChanged: this.topicSelectionChanged,
    };
    this.modalService.show(DashboardElementsOverlayComponent, { initialState });
  }
}
