import { Overlay, OverlayRef } from '@angular/cdk/overlay';
import { TemplatePortal } from '@angular/cdk/portal';
import {
  AfterViewInit,
  Component,
  ComponentRef,
  EventEmitter,
  OnDestroy,
  OnInit,
  Output,
  TemplateRef,
  ViewChild,
} from '@angular/core';
import { Subscription } from 'rxjs';
import { AbstractPage } from '../abstract-page/abstract-page';
import { ArticleCountComponent } from '../dashboard-elements/article-count/article-count.component';
import { HotMediaComponent } from '../dashboard-elements/hot-media/hot-media.component';
import { HotTopicsComponent } from '../dashboard-elements/hot-topics/hot-topics.component';
import { RegionalInfluenceComponent } from '../dashboard-elements/regional-influence/regional-influence.component';
import { TopicPopularityComponent } from '../dashboard-elements/topic-popularity/topic-popularity.component';
import { TopicReachComponent } from '../dashboard-elements/topic-reach/topic-reach.component';
import { DashboardElementDirective } from '../directives/dashboard-element.directive';
import { DashboardElement } from '../interfaces/dashboard-element';
import { OneUpApiService } from '../services/one-up-api.service';

/**
 *  defines the components selector and the components corresponding html and css file(s)
 */
@Component({
  selector: 'app-dashboard',
  templateUrl: './dashboard.component.html',
  styleUrls: ['./dashboard.component.css', '../app.component.css'],
  providers: [],
})
export class DashboardComponent extends AbstractPage implements AfterViewInit, OnInit, OnDestroy {
  /**
   *  TemplateRef to the overlay template that's shown when the backend is not reachable
   */
  @ViewChild('templateBackendNotReachable') templateBackendNotReachable: TemplateRef<unknown>;

  /**
   *  ng-template that is used to create a TemplatePortal, to which the overlay then attaches.
   */
  @ViewChild(DashboardElementDirective, { static: true }) dbElement: DashboardElementDirective;

  /**
   *  EventEmitter that fires when the user selected topics change
   */
  @Output() topicSelectionChanged: EventEmitter<any> = new EventEmitter<any>();

  /**
   *  Boolean indicator whether an overlay is already open.
   *  Prevents multiple overlays from being displayed on top of each other.
   */
  overlayIsOpen = false;

  /**
   *  The actual overlay reference, that is used to attach and detach the overlay.
   */
  overlayRef: OverlayRef;

  /**
   *  Portal that overlayRef attaches to.
   */
  templatePortal: TemplatePortal<any>;

  /**
   *  Subscribes to the backendNotReachable subscription to
   *  be notified when the backend is not reachable and no data can be retrieved
   */
  backendNotReachable: Subscription;

  /**
   *  Holds the current value of the colorScheme toggle. (deprecated)
   */
  toggleScheme = 'blue';

  /**
   *  Holds all visual parameters, in this case the color scheme
   */
  visOptions = {
    scheme: this.colorService.barCharts.blue,
  };

  /**
   *  Holds all angular components which are then added to the dashboard grid.
   */
  private dashboardFeatElement: any = {};

  /**
   *  If the query contains a valid 'feat' parameter, e.g. feat=articlecount
   *  only that component is added to the dashboard grid (instead of all components in dashboardFeatElement)
   */
  private feat = '';

  /**
   *  Inject the necessary services that are used in this component
   */
  constructor(private overlay: Overlay, private oneUpApiService: OneUpApiService) {
    super();
  }

  /**
   *  Defines which components are actually shown in the Dashboard and loads them.
   */
  ngOnInit(): void {
    this.route.queryParams.subscribe((params) => {
      if (params.feat !== undefined) {
        this.feat = params.feat;
      }
    });

    this.dashboardFeatElement = {
      articlecount: ArticleCountComponent,
      topreach: TopicReachComponent,
      toppop: TopicPopularityComponent,
      hottopic: HotTopicsComponent,
      hotmedia: HotMediaComponent,
      region: RegionalInfluenceComponent,
      // the following components are commented out for the evaluation with focus to science communication
      // article: ArticleTopicCircleComponent,
      // mensa: MensaComponent,
      // tdt: TdTComponent,
    };

    this.loadElements();
  }

  /**
   *  Subscribes to the backendNotReachable subscription, to also shown an overlay when the backend doesn't respond
   */
  ngAfterViewInit() {
    this.callAPIs();
  }

  /**
   *  Unsubscribe from subscriptions when page is destroyed
   */
  ngOnDestroy() {
    this.backendNotReachable.unsubscribe();
  }

  /**
   *  Subscribes to the backendNotReachable subscription, to also shown an overlay when the backend doesn't respond
   */
  callAPIs() {
    this.backendNotReachable = this.oneUpApiService.backendNotReachable.subscribe(
      (notReachable: boolean) => {
        if (notReachable) {
          this.templatePortal = new TemplatePortal(
            this.templateBackendNotReachable,
            this.dbElement.viewContainerRef
          );
          const positionStrategy = this.overlay
            .position()
            .global()
            .height('300px')
            .centerHorizontally()
            .centerVertically();
          this.overlayRef = this.overlay.create({
            positionStrategy,
            maxHeight: '90vh',
            hasBackdrop: true,
          });
          this.detachTemplate();
          this.attachOverlay();
        }
      }
    );
  }

  /**
   * Closes the overlay and returns to the Dashboard
   */
  detachTemplate() {
    if (this.overlayRef && this.overlayRef.hasAttached()) {
      this.overlayRef.detach();
    }
    this.overlayIsOpen = false;
  }

  /**
   * Emits an event when the user changes the color scheme.
   *
   * @param event - mat-button-toggle-group event that holds the newly selected color scheme
   */
  changedColorSelection(event) {
    this.toggleScheme = event.value;
    this.visOptions.scheme = this.colorService.barCharts[event.value];
    this.colorService.emitChangedColorSelection(event.value);
  }

  /**
   *  Attaches the overlay to the DOM.
   */
  private attachOverlay() {
    if (!this.overlayIsOpen) {
      this.overlayIsOpen = true;
      this.overlayRef.attach(this.templatePortal);
    }
  }

  /**
   *  Selects a random order of elements.
   *  Attaches all dashboard components to the viewContainer.
   *  If the query contains a valid 'feat' parameter, only that component is shown.
   */
  private loadElements() {
    const viewContainerRef = this.dbElement.viewContainerRef;
    viewContainerRef.clear();
    if (this.feat.length > 0 && this.dashboardFeatElement.hasOwnProperty(this.feat)) {
      const element = this.dashboardFeatElement[this.feat];
      const dashboardElement = viewContainerRef.createComponent<DashboardElement>(element);
      // all dashboardElements with an input of topicSelectionChanged get access to selected topics
      this.enhanceDashboardElement(dashboardElement);
    } else {
      for (const key in this.dashboardFeatElement) {
        if (this.dashboardFeatElement.hasOwnProperty(key)) {
          const dashboardElement = viewContainerRef.createComponent<DashboardElement>(
            this.dashboardFeatElement[key]
          );
          // all dashboardElements with an input of topicSelectionChanged get access to selected topics
          this.enhanceDashboardElement(dashboardElement);
        }
      }
    }
  }

  /**
   *  Adds the topicSelectionChanged EventEmitter to every dashboard component
   *  so that every component reacts to changes in the topic selection and updates its graphs.
   */
  private enhanceDashboardElement(dashboardElement: ComponentRef<DashboardElement>) {
    if (dashboardElement.instance.hasOwnProperty('topicSelectionChanged')) {
      // eslint-disable-next-line @typescript-eslint/dot-notation
      dashboardElement.instance['topicSelectionChanged'] = this.topicSelectionChanged;
    }
  }
}
