import { Overlay, OverlayRef } from '@angular/cdk/overlay';
import { TemplatePortal } from '@angular/cdk/portal';
import {
  Directive,
  ElementRef,
  HostListener,
  Input,
  OnDestroy,
  OnInit,
  TemplateRef,
  ViewContainerRef,
} from '@angular/core';
import { Subject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';
import { PopoverService } from './popover.service';

/**
 * Defines the popover directive.
 * Pass a template reference to the popover directive to display the template as a popover.
 */
@Directive({
  selector: '[appPopoverTrigger]',
})
export class PopoverDirective implements OnDestroy, OnInit {
  /**
   * The template reference to the popover content.
   */
  @Input() appPopoverTrigger!: TemplateRef<object>;

  /**
   * Whether the popover should close when the backdrop is clicked.
   */
  @Input() closeOnClickOutside = false;

  /**
   * Indicates when the directive is destroyed.
   */
  private unsubscribe = new Subject();

  /**
   * The overlay reference.
   */
  private overlayRef!: OverlayRef;

  /**
   * Inject the necessary services.
   */
  constructor(
    private elementRef: ElementRef,
    private overlay: Overlay,
    private vcr: ViewContainerRef,
    private popoverService: PopoverService
  ) {}

  /**
   * Attach the overlay when the host element
   * that uses the popover directive is clicked.
   */
  @HostListener('click') clickou() {
    this.attachOverlay();
  }

  /**
   * Initialize the overlay and subscribe to the popover service
   * to detach the overlay when the popover is closed.
   */
  ngOnInit(): void {
    this.createOverlay();
    this.popoverService.getState().subscribe((resp) => {
      if (resp) {
        this.detachOverlay();
      }
    });
  }

  /**
   * Unsubscribe from all subscriptions and
   * detach the overlay when the directive is destroyed.
   */
  ngOnDestroy(): void {
    this.detachOverlay();
    this.unsubscribe.next();
    this.unsubscribe.complete();
  }

  /**
   * Create the overlay and subscribe to the backdrop click event.
   */
  private createOverlay(): void {
    const scrollStrategy = this.overlay.scrollStrategies.block();
    const positionStrategy = this.overlay
      .position()
      .flexibleConnectedTo(this.elementRef)
      .withPositions([
        {
          originX: 'start',
          originY: 'bottom',
          overlayX: 'start',
          overlayY: 'bottom',
        },
      ]);

    this.overlayRef = this.overlay.create({
      positionStrategy,
      scrollStrategy,
      hasBackdrop: true,
      backdropClass: '',
    });

    this.overlayRef
      .backdropClick()
      .pipe(takeUntil(this.unsubscribe))
      .subscribe(() => {
        if (this.closeOnClickOutside) {
          this.detachOverlay();
        }
      });
  }

  /**
   * Attach the overlay to the DOM if it is not already attached.
   */
  private attachOverlay(): void {
    if (!this.overlayRef.hasAttached()) {
      const periodSelectorPortal = new TemplatePortal(this.appPopoverTrigger, this.vcr);

      this.overlayRef.attach(periodSelectorPortal);
    }
  }

  /**
   * Detach the overlay from the DOM if it is attached.
   */
  private detachOverlay(): void {
    if (this.overlayRef.hasAttached()) {
      this.overlayRef.detach();
    }
  }
}
