import {
  HttpEvent,
  HttpHandler,
  HttpInterceptor,
  HttpRequest,
  HttpResponse,
} from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Observable, of } from 'rxjs';
import { tap } from 'rxjs/operators';
import { HttpRequestCache } from '../services/http-request-cache.service';
import { OneUpApiService } from '../services/one-up-api.service';

/**
 * Cache interceptor stores the responses of API calls and use them when the API is called again.
 */
@Injectable()
export class CachingInterceptor implements HttpInterceptor {
  /**
   * Builds the Interceptor.
   *
   * @param cache The private cache is used to store and retrieve data.
   */
  constructor(private cache: HttpRequestCache, private apiService: OneUpApiService) {}

  /**
   * implemented function from HttpInterceptor
   * decides if the cache should be used
   *
   * @param req HttpRequest which could be cached
   * @param next next HttpHandler
   */
  intercept(req: HttpRequest<any>, next: HttpHandler) {
    // continue if not cacheable.
    if (!this.isCacheable(req)) {
      return next.handle(req);
    }

    const cachedResponse = this.cache.get(req);
    if (cachedResponse !== null) {
      this.apiService.resendCachedRequest(req.urlWithParams).subscribe((_) => {});
      return of(cachedResponse);
    }
    return this.sendRequest(req, next, this.cache);
  }

  /**
   * This function checks if the http request is cachable. It is not cachable if the param 'cached' is used.
   *
   * @param req HttpRequest
   */
  isCacheable(req: HttpRequest<any>) {
    const hasCached = req.urlWithParams.indexOf('cached') > -1;
    return !hasCached && !(req.method !== 'GET');
  }

  /**
   * Get server response observable by sending request to `next()`.
   * Will add the response to the cache on the way out.
   */
  sendRequest(
    req: HttpRequest<any>,
    next: HttpHandler,
    cache: HttpRequestCache
  ): Observable<HttpEvent<any>> {
    return next.handle(req).pipe(
      tap((event) => {
        // There may be other events besides the response.
        if (event instanceof HttpResponse) {
          cache.put(req, event); // Update the cache.
        }
      })
    );
  }
}
