/* eslint-disable @typescript-eslint/member-ordering */
/* eslint-disable no-underscore-dangle */
import { Injectable } from '@angular/core';
import * as d3 from 'd3';
import { BehaviorSubject } from 'rxjs';
import { convertToClassName } from '../util';

/**
 * Service that manages the highlighting of elements in the dashboard.
 */
@Injectable({
  providedIn: 'root',
})
export class HighlightService {
  /**
   * Emits the topic that is currently highlighted when the highlighting changes.
   */
  private _highlightedTopicChanged = new BehaviorSubject<string>('');

  /**
   * Emits the topic of the currently highlighted topic as an Observable.
   */
  highlightedTopicChanged = this._highlightedTopicChanged.asObservable();

  /**
   * Emits the color of the currently highlighted topic.
   */
  private _highlightedTopicColor = new BehaviorSubject<string>('');

  /**
   * Emits the color of the currently highlighted topic as an Observable.
   */
  highlightedTopicColor = this._highlightedTopicColor.asObservable();

  /**
   * Highlights all d3 elements with the given topic as class
   * by adding a red outline.
   *
   * @param topicToBeHighlighted the topic to be highlighted across the dashboard
   */
  highlightGraphs(topicToBeHighlighted: string) {
    const classToBeHighlighted = convertToClassName(topicToBeHighlighted);

    const clickedOnHighlightedClass =
      d3.select('.' + classToBeHighlighted).style('stroke-width') === '3px' ? true : false;

    this.removeHighlighting(clickedOnHighlightedClass ? false : true);

    if (!clickedOnHighlightedClass) {
      // highlight all d3 elements with selected class
      d3.selectAll('.' + classToBeHighlighted)
        .style('stroke-width', 3)
        .attr('stroke', 'red');

      this._highlightedTopicChanged.next(topicToBeHighlighted);
      this._highlightedTopicColor.next(d3.select('.' + classToBeHighlighted).style('fill'));
    }
  }

  /**
   * Removes highlighting from all d3 elements.
   *
   * @param suppressEventEmission if true, the event emission is suppressed
   */
  removeHighlighting(suppressEventEmission?: boolean) {
    // remove all current outlines
    d3.selectAll('rect').style('stroke-width', 1).attr('stroke', null);
    d3.selectAll('circle').style('stroke-width', 1).attr('stroke', null);

    if (!suppressEventEmission) {
      this._highlightedTopicChanged.next('');
    }
  }
}
