import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { BehaviorSubject, Observable, of } from 'rxjs';
import { catchError, map, take } from 'rxjs/operators';
import { environment } from '../../environments/environment';
import {
  HotTopic,
  MediaOverview,
  ReachOfTopic,
  TagsOverview,
  TopicWithArticle,
  TopMedium,
  UserTopic,
} from '../interfaces/one-up-api-response';
import { MessageService } from './message.service';

export enum MediaOverviewType {
  allTags = 'alltags',
  userTags = 'usertags',
  queryTag = 'querytag',
}

export enum TimespanType {
  lastMonth = 'lastMonth',
  lastThreeMonths = 'lastThreeMonths',
  totalPeriod = 'totalPeriod',
}

@Injectable({
  providedIn: 'root',
})
export class OneUpApiService {
  /* eslint-disable @typescript-eslint/member-ordering */
  /* eslint-disable no-underscore-dangle */
  private _backendNotReachable = new BehaviorSubject<boolean>(false);
  backendNotReachable = this._backendNotReachable.asObservable();

  constructor(private messageService: MessageService, private http: HttpClient) {}

  getTopics(user: string, mediaOverviewType: MediaOverviewType): Observable<TopicWithArticle[]> {
    const path =
      'articles/overview?mediaOverviewType=' + String(mediaOverviewType) + '&user=' + user;
    return this.http
      .get<TopicWithArticle[]>(environment.apiUrl + path, {
        observe: 'response',
      })
      .pipe(
        take(1),
        map((resp) => resp.body),
        catchError(this.handleError<TopicWithArticle[]>('getTopics', []))
      );
  }

  getTagsTimeview(
    mediaOverviewType: MediaOverviewType,
    query?: string
  ): Observable<ReachOfTopic[]> {
    let path = 'articles/tags/timeview?mediaOverviewType=' + String(mediaOverviewType);
    if (mediaOverviewType === MediaOverviewType.userTags) {
      path += '&user=' + query;
    } else if (mediaOverviewType === MediaOverviewType.queryTag) {
      path += '&tag=' + query;
    }
    return this.http.get<ReachOfTopic[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as ReachOfTopic[]),
      catchError(this.handleError<ReachOfTopic[]>('getTimeline', []))
    );
  }

  getHotTopics(user: string, mediaOverviewType: MediaOverviewType): Observable<HotTopic[]> {
    const path =
      'articles/top/tags?mediaOverviewType=' + String(mediaOverviewType) + '&user=' + user;
    return this.http.get<HotTopic[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as HotTopic[]),
      catchError(this.handleError<HotTopic[]>('getHotTopics', []))
    );
  }

  getHotMedia(
    mediaOverviewType: MediaOverviewType,
    selectedTimespan: TimespanType,
    query?: string,
    showAllMedia?: boolean
  ): Observable<TopMedium[]> {
    let path =
      'articles/top/media?mediaOverviewType=' +
      String(mediaOverviewType) +
      '&timespan=' +
      String(selectedTimespan);
    if (mediaOverviewType === MediaOverviewType.userTags) {
      path += '&user=' + query;
    } else if (mediaOverviewType === MediaOverviewType.queryTag) {
      path += '&tag=' + query;
    }

    if (showAllMedia) {
      path += '&showAllMedia=' + String(showAllMedia);
    }

    return this.http.get<TopMedium[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as TopMedium[]),
      catchError(this.handleError<TopMedium[]>('getHotMedia', []))
    );
  }

  getRecentArticles(
    mediaOverviewType: MediaOverviewType,
    amountOfArticles: number,
    query: string,
    mediumName?: string
  ): Observable<any> {
    let path =
      'articles/articles/medium/recent?mediaOverviewType=' +
      String(mediaOverviewType) +
      '&amountOfArticles=' +
      String(amountOfArticles) +
      '&medium_name=' +
      mediumName;
    if (mediaOverviewType === MediaOverviewType.userTags) {
      path += '&user=' + query;
    } else if (mediaOverviewType === MediaOverviewType.queryTag) {
      path += '&tag=' + query;
    }

    return this.http.get<any[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as any[]),
      catchError(this.handleError<any[]>('getRecentArticles', []))
    );
  }

  getUserTagsWithArticleCount(
    user: string,
    mediaOverviewType: MediaOverviewType,
    selectedTimespan: TimespanType
  ): Observable<MediaOverview[]> {
    const path =
      'articles/user/tags/article_count?mediaOverviewType=' +
      String(mediaOverviewType) +
      '&user=' +
      user +
      '&timespan=' +
      String(selectedTimespan);
    return this.http
      .get<any[]>(environment.apiUrl + path, {
        observe: 'response',
      })
      .pipe(
        take(1),
        map((response) => response.body as any[]),
        catchError(this.handleError<any[]>('getUserTagsWithArticleCount', []))
      );
  }

  getUserTopics(user: string): Observable<UserTopic[]> {
    const path = 'articles/user/tags?user=' + user;
    return this.http.get<UserTopic[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as UserTopic[]),
      catchError(this.handleError<UserTopic[]>('getUserTopics', [{ name: 'No User' }]))
    );
  }

  getTagsOverview(): Observable<TagsOverview[]> {
    const path = 'articles/tags/overview';
    return this.http.get<TagsOverview[]>(environment.apiUrl + path, { observe: 'response' }).pipe(
      take(1),
      map((response) => response.body as TagsOverview[]),
      catchError(this.handleError<TagsOverview[]>('getTagsOverview', []))
    );
  }

  postUserTags(user, tags): Observable<any> {
    const path = 'articles/user/tags';
    return this.http
      .post<any>(environment.apiUrl + path, { user, tags })
      .pipe(catchError(this.handleError<any>('postUserTags')));
  }

  postUserActivity(user: string, activity: string, object?: object): Observable<any> {
    const path = 'articles/user/activity?user=' + user + '&' + activity + '=true';
    return this.http
      .post(environment.apiUrl + path, object ? object : {})
      .pipe(catchError(this.handleError<any>('postUserActivity')));
  }

  resendCachedRequest(url: string): Observable<any[]> {
    const cachedUrl = url + '&cached=true';
    return this.http.get<any[]>(cachedUrl, { observe: 'response' }).pipe(
      take(1),
      map((response) => []),
      catchError(this.handleError<any[]>('resendCachedRequest', []))
    );
  }

  private handleError<T>(operation = 'operation', result?: T) {
    return (error: any): Observable<T> => {
      // TODO: send the error to remote logging infrastructure
      console.error(error); // log to console instead

      if (error.url === environment.apiUrl + 'articles/tags/overview') {
        this._backendNotReachable.next(true);
      }

      // TODO: better job of transforming error for user consumption
      this.messageService.log('MiSePR', `${operation} failed: ${error.message}`);

      // Let the app keep running by returning an empty result.
      return of(result as T);
    };
  }
}
