import { Injectable } from '@angular/core';
import * as d3 from 'd3';
import { BehaviorSubject } from 'rxjs';
import { convertToClassName } from 'src/app/util';
import { HighlightService } from '../highlight.service';
import { HoverTooltipService } from '../hover-tooltip.service';
import { ColorService } from './color.service';

@Injectable({
  providedIn: 'root',
})
export class BarchartService {
  /* eslint-disable @typescript-eslint/member-ordering */
  /* eslint-disable no-underscore-dangle */
  private _changeMediumDetailEmitter = new BehaviorSubject<string>('');
  changeMediumDetailEmitter = this._changeMediumDetailEmitter.asObservable();

  constructor(
    private colorService: ColorService,
    private highlightService: HighlightService,
    private hoverTooltipService: HoverTooltipService
  ) {}

  createSVGMedia(element, width, height, margin) {
    return d3
      .select(element)
      .append('svg')
      .attr('width', width + margin.left + margin.right)
      .attr('height', height + margin.top + margin.bottom)
      .append('g')
      .attr('transform', 'translate(' + margin.left + ',' + margin.top + ')');
  }

  createAxisSVG(element, visOptions) {
    return d3
      .select(element)
      .append('svg')
      .attr('width', visOptions.width + visOptions.margin.left + visOptions.margin.right)
      .attr('height', visOptions.height + visOptions.margin.top + visOptions.margin.bottom)
      .style('position', 'absolute')
      .style('pointer-events', 'none')
      .style('z-index', 1)
      .append('g')
      .attr('transform', 'translate(' + visOptions.margin.left + ',' + visOptions.margin.top + ')');
  }

  createBody(element, visOptions) {
    if (visOptions.scrollDirection === 'vertical') {
      return d3
        .select(element)
        .append('div')
        .style('-webkit-overflow-scrolling', 'touch')
        .style('overflow-y', 'scroll')
        .style('overflow-x', 'hidden')
        .style('max-height', `${visOptions.height + 10}px`)
        .style('max-width', `${visOptions.width + 10}px`);
    } else if (visOptions.scrollDirection === 'horizontal') {
      return d3
        .select(element)
        .append('div')
        .style('-webkit-overflow-scrolling', 'touch')
        .style('overflow-x', 'scroll')
        .style('max-width', `${visOptions.width + 10}px`)
        .style('margin-left', '60px');
    }
  }

  selectBar(event, d, isOverlayActive, tooltip) {
    if (d.tag_name) {
      this.highlightService.highlightGraphs(d.tag_name);
    } else if (d.medium_name) {
      this.hoverTooltipService.mouseleave(event, d, tooltip);
      this._changeMediumDetailEmitter.next(
        JSON.stringify({ mediumName: d.medium_name, activeOverlay: isOverlayActive })
      );
    }
  }

  createUpdateHorizontalBars(options, data, seperateXAxis?) {
    const sortedData = data.sort((a, b) =>
      d3.descending(a[options.xColumnName], b[options.xColumnName])
    );
    const transition = options.element
      .transition()
      .duration(800)
      .delay((d, i) => i * 100);

    // set the values for the corresponding axis
    const x = d3.scaleLinear().range([0, options.width]).domain([0, options.maxRange]);

    // add y axis
    const y = d3
      .scaleBand()
      .range(seperateXAxis ? [0, options.height * 6] : [0, options.height])
      .domain(sortedData.map((d) => d[options.yColumnName]))
      .padding(1 / data.length > 0.1 ? 1 / data.length : 0.1);

    if (seperateXAxis) {
      options.element.select('#svgBody').remove();
      options.body = options.element
        .append('svg')
        .attr('id', 'svgBody')
        .attr('width', options.width + options.margin.left + options.margin.right)
        .attr('height', (data.length + 2) * (y.bandwidth() * 1.104))
        .style('display', 'block')
        .append('g')
        .attr('transform', 'translate(' + 0 + ',' + options.margin.top + ')');
    }

    const axisToAppendTo = seperateXAxis ? options.xAxis : options.element;
    const bodyToAppendTo = seperateXAxis ? options.body : options.element;

    axisToAppendTo.select('#xAxis').remove();
    axisToAppendTo.select('#xAxisLabel').remove();
    bodyToAppendTo.select('#yAxis').remove();

    axisToAppendTo
      .append('text')
      .attr('id', 'xAxisLabel')
      .attr('class', 'x label')
      .attr('text-anchor', 'middle')
      .attr('x', options.width / 2)
      .attr('y', options.height + 40)
      .style('font-size', '14px')
      .text(options.xTitle);

    // create a tooltip
    const tooltip = this.hoverTooltipService.createTooltip();

    // add bars
    bodyToAppendTo
      .selectAll('rect')
      .data(sortedData)
      .join(
        (enter) =>
          enter
            .append('rect')
            .style('cursor', 'pointer')
            .attr('x', x(0))
            .attr('y', (d) => y(d[options.yColumnName]))
            .attr('height', y.bandwidth())
            .attr('fill', (d) => options.scheme(d[options.yColumnName]))
            .style('opacity', '0.65')
            // no bars at beginning
            .attr('width', (d) => 0)
            .attr('class', (d) =>
              d[options.yColumnName] ? convertToClassName(d[options.yColumnName]) : ''
            )
            .on('click', (event, d) => this.selectBar(event, d, seperateXAxis, tooltip))
            .on('mouseover', (event, d) => this.hoverTooltipService.mouseover(event, d, tooltip))
            .on('mousemove', (event, d) =>
              this.hoverTooltipService.mousemove(
                event,
                d,
                tooltip,
                d[options.yColumnName] + '<br>' + options.xTitle + ': ' + d[options.xColumnName]
              )
            )
            .on('mouseleave', (event, d) => this.hoverTooltipService.mouseleave(event, d, tooltip))
            .call((enter) =>
              enter
                .transition(transition)
                .attr('x', x(0))
                .attr('width', (d) => x(d[options.xColumnName]))
            ),
        (update) =>
          update
            .attr('fill', this.colorService.loading.primary)
            .style('cursor', 'pointer')
            .attr('class', (d) =>
              d[options.yColumnName] ? convertToClassName(d[options.yColumnName]) : ''
            )
            .call((update) =>
              update
                .transition(transition)
                .attr('x', x(0))
                .attr('y', (d) => y(d[options.yColumnName]))
                .attr('width', (d) => x(d[options.xColumnName]))
                .attr('height', y.bandwidth())
                .attr('fill', (d) => options.scheme(d[options.yColumnName]))
            ),
        (exit) => exit.call((exit) => exit.transition(transition).remove())
      );

    // add x axis labels
    axisToAppendTo
      .append('g')
      .attr('transform', 'translate(0,' + options.height + ')')
      .attr('id', 'xAxis')
      .call(d3.axisBottom(x))
      .selectAll('text')
      .attr('transform', 'translate(-10,0)rotate(-45)')
      .style('text-anchor', 'end');

    const yAxis = (g) =>
      g
        .attr('transform', `translate(${x(0)},0)`)
        .attr('id', 'yAxis')
        .call(d3.axisRight(y).tickSizeOuter(0));
    bodyToAppendTo.append('g').call(yAxis);
  }

  createUpdateVerticalBars(options, data, userFavoriteTopics, seperateYAxis?) {
    const x = d3
      .scaleBand()
      .range(seperateYAxis ? [0, options.width * 2] : [0, options.width])
      .domain(data.map((d) => d[options.xColumnName]))
      .padding(seperateYAxis ? (1 / data.length > 0.1 ? 1 / data.length : 0.1) : 0.6 / data.length);

    const transition = options.element.transition().duration(800);

    if (seperateYAxis) {
      options.element.select('#svgBody').remove();
      options.body = options.element
        .append('svg')
        .attr('id', 'svgBody')
        .attr('width', (data.length + 2) * (x.bandwidth() * 1.1))
        .attr('height', options.height + options.margin.top + options.margin.bottom)
        .style('display', 'block')
        .append('g')
        .attr('transform', 'translate(' + 0 + ',' + options.margin.top + ')');
    }

    const axisToAppendTo = seperateYAxis ? options.yAxis : options.element;
    const bodyToAppendTo = seperateYAxis ? options.body : options.element;

    axisToAppendTo.select('#xAxis').remove();
    axisToAppendTo.select('#yAxisLabel').remove();
    axisToAppendTo.select('#yAxis').remove();

    const y = d3.scaleLinear().range([options.height, 0]).domain([0, data[0][options.yColumnName]]);

    const yAxis = (g) => g.attr('id', 'yAxis').call(d3.axisLeft(y));
    axisToAppendTo.append('g').call(yAxis);

    axisToAppendTo
      .append('text')
      .attr('id', 'yAxisLabel')
      .attr('transform', 'rotate(-90)')
      .attr('text-anchor', 'middle')
      .attr('x', 0 - options.height / 2)
      .attr('y', 0 - options.margin.left)
      .attr('dy', '1em')
      .style('font-size', '14px')
      .text(options.xTitle);

    // create a tooltip
    const tooltip = this.hoverTooltipService.createTooltip();

    bodyToAppendTo
      .selectAll('rect')
      .data(data)
      .join(
        (enter) =>
          enter
            .append('rect')
            .style('cursor', 'pointer')
            .attr('y', y(0))
            .attr('height', 0)
            .attr('fill', (d) =>
              this.isUserFavorite(userFavoriteTopics, d)
                ? options.scheme(d[options.xColumnName])
                : '#D7D7D7'
            )
            .style('opacity', '0.65')
            .attr('x', (d) => x(d[options.xColumnName]))
            // no bars at beginning
            .attr('width', x.bandwidth())
            .attr('class', (d) =>
              d[options.xColumnName] ? convertToClassName(d[options.xColumnName]) : ''
            )
            .on('click', (event, d) => this.selectBar(event, d, seperateYAxis, tooltip))
            .on('mouseover', (event, d) => this.hoverTooltipService.mouseover(event, d, tooltip))
            .on('mousemove', (event, d) =>
              this.hoverTooltipService.mousemove(
                event,
                d,
                tooltip,
                d[options.xColumnName] + '<br>' + options.xTitle + ': ' + d[options.yColumnName]
              )
            )
            .on('mouseleave', (event, d) => this.hoverTooltipService.mouseleave(event, d, tooltip))
            .call((enter) =>
              enter
                .transition(transition)
                .attr('y', (d) => y(d[options.yColumnName]))
                .attr('height', (d) => y(0) - y(d[options.yColumnName]))
            ),
        (update) =>
          update
            .attr('fill', this.colorService.loading.primary)
            .style('cursor', 'pointer')
            .attr('class', (d) =>
              d[options.xColumnName] ? convertToClassName(d[options.xColumnName]) : ''
            )
            .call((update) =>
              update
                .transition(transition)
                .attr('x', (d) => x(d[options.xColumnName]))
                .attr('width', x.bandwidth())
                .attr('y', (d) => y(d[options.yColumnName]))
                .attr('height', (d) => y(0) - y(d[options.yColumnName]))
                .attr('fill', (d) =>
                  this.isUserFavorite(userFavoriteTopics, d)
                    ? options.scheme(d[options.xColumnName])
                    : '#D7D7D7'
                )
            ),
        (exit) => exit.call((exit) => exit.transition(transition).remove())
      );

    const xAxis = bodyToAppendTo
      .append('g')
      .attr('id', 'xAxis')
      .attr('transform', 'translate(0,' + options.height + ')');

    xAxis
      .call(d3.axisTop(x).tickSizeOuter(0))
      .selectAll('text')
      .attr('transform', 'rotate(-90)')
      .attr('x', 10)
      .attr('y', 3)
      .style('text-anchor', 'start');
  }

  createUpdatePyramidBars(options, realData, userFavoriteTopics, seperateXAxis?) {
    let data = JSON.parse(JSON.stringify(realData)); // we need this deep copy because otherwise the cache would be overwritten
    data.sort((a, b) => d3.descending(a[options.xColumnLeftName], b[options.xColumnLeftName]));

    const maxRangeLeft = d3.max(data, (d) => Math.abs(d[options.xColumnLeftName]));
    const maxRangeRight = d3.max(data, (d) => Math.abs(d[options.xColumnRightName]));
    const maxRange = Math.max(maxRangeLeft, maxRangeRight);

    // filter out items with values less than 1% of the scale, because they'd be invisible
    //data = data.filter((item) => Number(item[options.xColumnLeftName]) > maxRange / 100);

    const newData = [];
    for (const element of data) {
      if (options.xColumnRightName !== options.xColumnLeftName) {
        const tmp = Object.assign({}, element);
        delete element[options.xColumnRightName];
        newData.push(element);
        delete tmp[options.xColumnLeftName];
        newData.push(tmp);
      } else {
        newData.push(element);
      }
    }

    data = newData;

    const multiplier = options.negativeScale ? -1 : 1;
    const xLeft = d3
      .scaleLinear()
      .domain([0, maxRange * 1.2 * multiplier])
      .range([options.width / 2, 0]);

    const xRight = d3
      .scaleLinear()
      .domain([0, maxRange * 1.2])
      .range([options.width / 2, options.width]);

    const y = d3
      .scaleBand()
      .domain(data.map((d) => d[options.yColumnName]))
      .range(seperateXAxis ? [0, options.height * 4] : [0, options.height])
      .padding(
        1 / data.length > 0.1
          ? options.xColumnRightName !== options.xColumnLeftName
            ? 1 / data.length
            : 1 / (data.length * 2)
          : 0.1
      );

    if (seperateXAxis) {
      options.element.select('#svgBody').remove();
      options.body = options.element
        .append('svg')
        .attr('id', 'svgBody')
        .attr('width', options.width + options.margin.left + options.margin.right)
        .attr(
          'height',
          options.xColumnRightName !== options.xColumnLeftName
            ? (data.length / 2 + 1) * (y.bandwidth() * 1.107)
            : (data.length / 2 + 1) * (y.bandwidth() * 2.197)
        )
        .style('display', 'block')
        .append('g')
        .attr('transform', 'translate(' + options.margin.left + ',' + options.margin.top + ')');
    }

    const xAxis = (g) =>
      g
        .attr('transform', `translate(0,${options.height})`)
        .attr('id', 'xAxis')
        .call((g) => g.append('g').call(d3.axisBottom(xLeft).ticks(options.width / 80, 's')))
        .call((g) => g.append('g').call(d3.axisBottom(xRight).ticks(options.width / 80, 's')))
        .call((g) => g.selectAll('.tick:first-of-type').remove());

    const yAxis = (g) =>
      g
        .attr('transform', `translate(${xRight(0)},0)`)
        .attr('id', 'yAxis')
        .call(d3.axisRight(y).tickSizeOuter(0));

    const elementToAppendTo = seperateXAxis ? options.xAxis : options.element;
    const bodyToAppendTo = seperateXAxis ? options.body : options.element;
    elementToAppendTo.select('#xAxis').remove();
    elementToAppendTo.select('#xTitle').remove();
    elementToAppendTo.select('#xTitleLeft').remove();
    elementToAppendTo.select('#xTitleRight').remove();
    bodyToAppendTo.select('#yAxis').remove();

    const transition = bodyToAppendTo.transition().duration(800);

    // create a tooltip
    const tooltip = this.hoverTooltipService.createTooltip();

    bodyToAppendTo
      .selectAll('rect')
      .data(data)
      .join(
        (enter) =>
          enter
            .append('rect')
            .style('cursor', 'pointer')
            .attr('class', (d) =>
              d[options.yColumnName] ? convertToClassName(d[options.yColumnName]) : ''
            )
            .attr('x', (d) => xRight(0))
            .attr('y', (d) => y(d[options.yColumnName]))
            .attr('height', y.bandwidth())
            .attr('fill', (d) =>
              this.isUserFavorite(userFavoriteTopics, d)
                ? options.scheme(d[options.yColumnName])
                : '#D7D7D7'
            )
            .style('opacity', (d) =>
              d[options.xColumnLeftName] && d[options.xColumnLeftName] > 0 ? '0.85' : '0.65'
            )
            .on('click', (event, d) => this.selectBar(event, d, seperateXAxis, tooltip))
            .on('mouseover', (event, d) => this.hoverTooltipService.mouseover(event, d, tooltip))
            .on('mousemove', (event, d) =>
              this.hoverTooltipService.mousemove(
                event,
                d,
                tooltip,
                d[options.yColumnName] +
                  '<br>' +
                  options.xTitle +
                  (options.xTitleLeft && d[options.xColumnLeftName]
                    ? ' (' + options.xTitleLeft + ')'
                    : options.xTitleRight && d[options.xColumnRightName]
                    ? ' (' + options.xTitleRight + ')'
                    : '') +
                  ': ' +
                  (d[options.xColumnLeftName]
                    ? d[options.xColumnLeftName]
                    : d[options.xColumnRightName])
              )
            )
            .on('mouseleave', (event, d) => this.hoverTooltipService.mouseleave(event, d, tooltip))
            .call((enter) =>
              enter
                .transition(transition)
                .attr('x', (d) =>
                  (d[options.xColumnLeftName] &&
                    options.xColumnLeftName !== options.xColumnRightName) ||
                  (d[options.xColumnRightName] && d[options.xColumnRightName] < 0)
                    ? xLeft(d[options.xColumnLeftName])
                    : xRight(0)
                )
                .attr('width', (d) =>
                  (d[options.xColumnLeftName] &&
                    options.xColumnLeftName !== options.xColumnRightName) ||
                  (d[options.xColumnRightName] && d[options.xColumnRightName] < 0)
                    ? xLeft(0) - xLeft(d[options.xColumnLeftName])
                    : xRight(d[options.xColumnRightName]) - xRight(0)
                )
            ),
        (update) =>
          update
            .attr('fill', this.colorService.loading.primary)
            .attr('class', (d) =>
              d[options.yColumnName] ? convertToClassName(d[options.yColumnName]) : ''
            )
            .call((update) =>
              update
                .transition(transition)
                .attr('x', (d) =>
                  (d[options.xColumnLeftName] &&
                    options.xColumnLeftName !== options.xColumnRightName) ||
                  (d[options.xColumnRightName] && d[options.xColumnRightName] < 0)
                    ? xLeft(d[options.xColumnLeftName])
                    : xRight(0)
                )
                .attr('y', (d) => y(d[options.yColumnName]))
                .attr('width', (d) =>
                  (d[options.xColumnLeftName] &&
                    options.xColumnLeftName !== options.xColumnRightName) ||
                  (d[options.xColumnRightName] && d[options.xColumnRightName] < 0)
                    ? xLeft(0) - xLeft(d[options.xColumnLeftName])
                    : xRight(d[options.xColumnRightName]) - xRight(0)
                )
                .attr('height', y.bandwidth())
                .attr('fill', (d) =>
                  this.isUserFavorite(userFavoriteTopics, d)
                    ? options.scheme(d[options.yColumnName])
                    : '#D7D7D7'
                )
            ),
        (exit) => exit.call((exit) => exit.transition(transition).remove())
      );

    elementToAppendTo
      .append('text')
      .attr('id', 'xTitleLeft')
      .attr('text-anchor', 'start')
      .attr('x', 0)
      .attr('y', options.height - 6)
      .style('font-size', '14px')
      .text(options.xTitleLeft);

    elementToAppendTo
      .append('text')
      .attr('id', 'xTitleRight')
      .attr('text-anchor', 'end')
      .attr('x', options.width)
      .attr('y', options.height - 6)
      .style('font-size', '14px')
      .text(options.xTitleRight);

    elementToAppendTo
      .append('text')
      .attr('id', 'xTitle')
      .attr('class', 'x label')
      .attr('text-anchor', 'start')
      .attr('x', options.width / 2 - options.xTitle.length * 3)
      .attr('y', options.height + 35)
      .style('font-size', '14px')
      .text(options.xTitle);

    elementToAppendTo.append('g').call(xAxis);

    bodyToAppendTo.append('g').call(yAxis);
  }

  isUserFavorite(userFavoriteTopics, topic): boolean {
    let isFavorite = false;
    for (const favoriteTopic of userFavoriteTopics) {
      if (favoriteTopic.name === topic.tag_name) {
        isFavorite = true;
        break;
      }
    }
    return isFavorite;
  }
}
