import { Injectable } from '@angular/core';
import * as d3 from 'd3';
import { TopicWithArticle } from 'src/app/interfaces/one-up-api-response';
import { convertToClassName } from 'src/app/util';
import { HighlightService } from '../highlight.service';
import { ColorService } from './color.service';

/**
 * This service creates a visualization with circles to show how many articles are attached to one tag.
 * [Model for the visualization](https://observablehq.com/@d3/zoomable-circle-packing)
 */
@Injectable({
  providedIn: 'root',
})
export class CircleGraphService {
  /**
   * A function which is injected by the view and is also handled by the view.
   * If nothing is injected then nothing happens.
   */
  private seeDetailsFunction;

  /**
   * The constructor of the service.
   *
   * @param colorService a service which is responsible for the color scheme
   * @param highlightService A service that provides information about a highlighted topic. Can be used to highlight or to give a signal that it should be highlighted.
   */
  constructor(private colorService: ColorService, private highlightService: HighlightService) {}

  /**
   * creates the component in which the diagram will be drawn
   *
   * @param element html component in which the svg will be created
   * @param options options for sizing the components (needed attributes width, height, margin.top, margin.bottom)
   * @param expandedView true if the svg is used for the overlay
   */
  createSVG(element, options, expandedView?) {
    return d3
      .select(element)
      .append('svg')
      .attr(
        'viewBox',
        `-${options.width / 2} -${
          options.height / 2 + options.margin.top + options.margin.bottom
        } ${options.width} ${options.height}`
      )
      .style('margin', expandedView ? '-15px' : '-45px')
      .style('display', 'block')
      .style('background', 'white');
  }

  /**
   * creates a circle graph
   *
   * @param svg the graph is built in this svg
   * @param height height of the svg
   * @param width width of the svg
   * @param data data to plot
   * @param colorScheme color scheme in which the circles are drawn
   * @param userFavoriteTopics highlighted topic
   * @param expandedView  true if the svg is used for the overlay
   */
  createChart(
    svg,
    height: number,
    width: number,
    data,
    colorScheme,
    userFavoriteTopics,
    expandedView?
  ) {
    data = this.convertToChartData(data);

    const t = svg
      .transition()
      .duration(800)
      .delay((d, i) => i * 100);
    const pack = (data) =>
      d3.pack().size([width, height]).padding(3)(
        d3
          .hierarchy(data)
          .sum((d) => d.value)
          .sort((a, b) => b.value - a.value)
      );
    const root = pack(data);

    svg.select('#titlePaths').remove();
    const paths = svg
      .append('defs')
      .attr('id', 'titlePaths')
      .selectAll('path')
      .data(root.descendants())
      .join('path')
      .attr('id', (d) =>
        expandedView ? 'curvedLabelPathExpanded' + d.data.name : 'curvedLabelPath' + d.data.name
      )
      .attr('d', (d) =>
        d3
          .arc()
          .innerRadius(d.r * 0.9 + 2.5)
          .outerRadius(d.r * 0.9 + 2.5)
          .startAngle(-Math.PI)
          .endAngle(Math.PI)()
      );

    // reduces the amount of child-circles to 10, if the overlay is present
    // otherwise the performance is too bad
    if (expandedView) {
      for (const parent in root.children) {
        if (Object.prototype.hasOwnProperty.call(root.children, parent)) {
          root.children[parent].children = root.children[parent].children.slice(0, 10);
        }
      }
    }

    const node = svg
      .selectAll('circle')
      .data(root.descendants().slice(1))
      .join(
        (enter) =>
          enter
            .append('circle')
            .style('cursor', 'pointer')
            .attr('fill', (d) => {
              if (d.children) {
                if (this.isUserFavorite(userFavoriteTopics, d)) {
                  return colorScheme(d.data.name);
                } else {
                  return '#D7D7D7';
                }
              } else {
                return 'white';
              }
            })
            .style('opacity', (d) => (d.depth === 1 ? '0.65' : '1.0'))
            .attr('pointer-events', (d) => (!d.children ? 'none' : null))
            .attr('class', (d) => (d.depth === 1 ? convertToClassName(d.data.name) : ''))
            .on('mouseover', function () {
              d3.select(this).attr('stroke', '#000');
            })
            .on('mouseout', function () {
              if (d3.select(this).style('stroke-width') === '3px') {
                d3.select(this).attr('stroke', 'red');
              } else {
                d3.select(this).attr('stroke', null);
              }
            })
            .on('click', (event, d) => this.seeDetails(d))
            .call((enter) => enter.transition(t)),
        (update) =>
          update
            .attr('fill', (d) =>
              d.children ? this.colorService.loading.primary : this.colorService.loading.secondary
            )
            .style('cursor', 'pointer')
            .attr('pointer-events', (d) => (!d.children ? 'none' : null))
            .attr('class', (d) => (d.depth === 1 ? convertToClassName(d.data.name) : ''))
            .on('mouseover', function () {
              d3.select(this).attr('stroke', '#000');
            })
            .on('mouseout', function () {
              if (d3.select(this).style('stroke-width') === '3px') {
                d3.select(this).attr('stroke', 'red');
              } else {
                d3.select(this).attr('stroke', null);
              }
            })
            .on('click', (event, d) => this.seeDetails(d))
            .call((update) =>
              update.transition(t).attr('fill', (d) => {
                if (d.children) {
                  if (this.isUserFavorite(userFavoriteTopics, d)) {
                    return colorScheme(d.data.name);
                  } else {
                    return '#D7D7D7';
                  }
                } else {
                  return 'white';
                }
              })
            ),
        (exit) => exit.call((exit) => exit.transition(t).remove())
      );

    svg.select('#titles').remove();
    const label = svg
      .append('g')
      .attr('id', 'titles')
      .style('font', '8px sans-serif')
      .attr('pointer-events', 'none')
      .selectAll('text')
      .data(root.descendants())
      .join('text')
      .append('textPath')
      .attr('class', 'curved-labels')
      .attr('href', (d) =>
        expandedView ? '#curvedLabelPathExpanded' + d.data.name : '#curvedLabelPath' + d.data.name
      )
      .attr('text-anchor', 'middle')
      .attr('startOffset', '25%')
      .style('fill-opacity', (d) => (d.parent === root ? 1 : 0))
      .style('display', (d) => (d.parent === root ? 'inline' : 'none'))
      .text((d) => d.data.name)
      .each(function (d, i) {
        // truncate labels that are longer than the circle's circumference
        if (d.parent === root) {
          const circum = 2 * Math.PI * d.r;
          let textWidth = this.getComputedTextLength();
          let text = d.data.name;

          while (textWidth > circum - circum / 10) {
            text = text.slice(0, -6) + '...';
            d3.select(this).text(text);
            textWidth = this.getComputedTextLength();
          }
        }
      });

    this.zoomTo([root.x, root.y + 20, root.r * 2], label, node, paths);
  }

  /**
   * highlight node
   *
   * @param nodeName highlighted node
   */
  selectCircle(nodeName) {
    if (nodeName) {
      this.highlightService.highlightGraphs(nodeName);
    }
  }

  /**
   * topics with articles were converted in a parent (topic) child (article) format
   *
   * @param twas articles which are converted
   */
  convertToChartData(twas: TopicWithArticle[]) {
    const res = {
      name: 'head',
      children: [],
    };
    for (const twa of twas) {
      let newChild = {
        name: '',
        children: [],
      };
      const indexOfTag = res.children.findIndex((el) => el.name === twa.name);
      if (indexOfTag >= 0) {
        newChild = res.children[indexOfTag];
      } else {
        newChild.name = twa.name;
        res.children.push(newChild);
      }
      const article = {
        name: twa.article_name,
        value: twa.reach,
      };
      newChild.children.push(article);
    }
    return res;
  }

  /**
   * computes if a topic is a favorite topic of the user
   *
   * @param userFavoriteTopics favorite topics of the user
   * @param topic all possible topics
   */
  isUserFavorite(userFavoriteTopics, topic): boolean {
    let isFavorite = false;
    for (const favoriteTopic of userFavoriteTopics) {
      if (favoriteTopic.name === topic.data.name) {
        isFavorite = true;
        break;
      }
    }
    return isFavorite;
  }

  /**
   * show details if the circle is clicked
   *
   * @param node clicked circle
   * @private
   */
  private seeDetails(node) {
    this.selectCircle(node.data.name);
    if (this.seeDetailsFunction !== undefined) {
      this.seeDetailsFunction(node.data.name);
    }
  }

  /**
   * zoom to a specific view port
   *
   * @param v middle of the view port [x, y, radius]
   * @param label label which is written to the specific circle
   * @param node zoomed to node
   * @param paths curved label paths
   * @private
   */
  private zoomTo(v, label, node, paths) {
    paths.attr('transform', this.translation(v));
    label.attr('transform', this.translation(v));
    node.attr('transform', this.translation(v));
    node.attr('r', (d) => d.r * 0.9);
  }

  /**
   * helper method to create same translation in paths, label and node
   *
   * @param v viewport
   * @private
   */
  private translation(v) {
    return (d) => `translate(${(d.x - v[0]) * 0.9},${(d.y - v[1]) * 0.9})`;
  }
}
