import { Injectable } from '@angular/core';
import * as d3 from 'd3';
import { ReachOfTopic } from 'src/app/interfaces/one-up-api-response';
import { HoverTooltipService } from '../hover-tooltip.service';
// import { convertToClassName } from 'src/app/util';
import { ColorService } from './color.service';

const localeDE = {
  dateTime: '%A, der %e. %B %Y, %X',
  date: '%d.%m.%Y',
  time: '%H:%M:%S',
  periods: ['AM', 'PM'],
  days: ['Sonntag', 'Montag', 'Dienstag', 'Mittwoch', 'Donnerstag', 'Freitag', 'Samstag'],
  shortDays: ['So', 'Mo', 'Di', 'Mi', 'Do', 'Fr', 'Sa'],
  months: [
    'Januar',
    'Februar',
    'März',
    'April',
    'Mai',
    'Juni',
    'Juli',
    'August',
    'September',
    'Oktober',
    'November',
    'Dezember',
  ],
  shortMonths: [
    'Jan',
    'Feb',
    'März',
    'Apr',
    'Mai',
    'Jun',
    'Jul',
    'Aug',
    'Sep',
    'Okt',
    'Nov',
    'Dez',
  ],
};

@Injectable({
  providedIn: 'root',
})
export class LineGraphService {
  constructor(
    private colorService: ColorService,
    private hoverTooltipService: HoverTooltipService
  ) {}

  createSVG(element, options) {
    return d3
      .select(element)
      .append('svg')
      .attr('width', options.width + options.margin.left + options.margin.right)
      .attr('height', options.height + options.margin.top + options.margin.bottom)
      .append('g')
      .attr('transform', 'translate(' + options.margin.left + ',' + options.margin.top + ')');
  }

  createUpdateLineGraph(options, mappedData, userFavoriteTopics) {
    mappedData = this.convertToChartData(mappedData);

    options.element.select('#xAxis').remove();
    options.element.selectAll('#line').remove();
    options.element.select('#yAxis').remove();

    d3.timeFormatDefaultLocale(localeDE);

    let allValues = [];
    mappedData.forEach((value, _) => (allValues = allValues.concat(value)));

    const xScale = d3
      .scaleTime()
      .range([0, options.width])
      .domain(d3.extent(allValues, (d) => d.date));

    const yScale = d3
      .scaleLinear()
      .rangeRound([options.height, 0])
      .domain([0, d3.max(mappedData, (c) => d3.max(c[1], (d) => d[options.yColumnName] + 4))]);

    const yAxis = d3
      .axisLeft()
      .scale(yScale)
      .tickFormat((d) => (d >= 10 ? d3.format('.2s')(d) : d));

    const xAxis = d3
      .axisBottom()
      .scale(xScale)
      .ticks(options.width > 1000 ? options.width / 120 : options.width / 100)
      .tickFormat(d3.timeFormat("%B '%y"));

    options.element
      .append('g')
      .attr('id', 'xAxis')
      .attr('transform', 'translate(0,' + options.height + ')')
      .call(xAxis);

    options.element.append('g').attr('id', 'yAxis').call(yAxis);

    const line = d3
      .line()
      .x((d) => xScale(d.date))
      .y((d) => yScale(d[options.yColumnName]));

    // create a tooltip
    const tooltip = this.hoverTooltipService.createTooltip();

    const bisect = (d, pointer) => {
      // get the x-value of the mouse pointer
      const x0 = xScale.invert(pointer[0]);

      // This allows to find the closest X index of the mouse:
      const bisectX = d3.bisector((d) => d.date).center;

      // determine which data point is closest to the pointers x-value
      let i = bisectX(d, x0, 1);

      // make sure the index is never out of bounds
      i = i >= d.length ? d.length - 1 : i < 0 ? 0 : i;
      const selectedData = d[i];
      const option = {
        // day: '2-digit',
        month: 'long',
        year: 'numeric',
      };

      return (
        selectedData[options.yColumnName] +
        ' (' +
        selectedData[options.xColumnName].toLocaleDateString('de-DE', option) +
        ')'
      );
    };

    const lines = options.element.selectAll('lines').data(mappedData).enter().append('g');
    lines
      .append('path')
      .attr('d', (d) => line(d[1]))
      .attr('id', 'line')
      // .attr('class', (d) => (d[0] ? convertToClassName(d[0]) : ''))
      .attr('fill', 'none')
      .attr('stroke', (d) =>
        this.isUserFavorite(userFavoriteTopics, d) ? options.scheme(d[0]) : '#c8c8c8'
      )
      .attr('stroke-width', (d) => (this.isUserFavorite(userFavoriteTopics, d) ? '3px' : '0.5px'))
      .on('mouseover', (event, d) => this.hoverTooltipService.mouseover(event, d, tooltip))
      .on('mousemove', (event, d) =>
        this.hoverTooltipService.mousemove(
          event,
          d,
          tooltip,
          d[0] +
            '<br>' +
            (options.yColumnName === 'articleCount' ? 'Artikelanzahl' : 'Reichweite') +
            ': ' +
            bisect(d[1], d3.pointer(event, event.target))
        )
      )
      .on('mouseleave', (event, d) => this.hoverTooltipService.mouseleave(event, d, tooltip));
  }

  isUserFavorite(userFavoriteTopics, topic): boolean {
    let isFavorite = false;
    for (const favoriteTopic of userFavoriteTopics) {
      if (favoriteTopic.name === topic[0]) {
        isFavorite = true;
        break;
      }
    }
    return isFavorite;
  }

  convertToChartData(data): Map<string, ReachOfTopic[]> {
    const timeRecord = new Map<string, any[]>();
    for (const tag of data) {
      if (!timeRecord.has(tag.name)) {
        timeRecord.set(tag.name, []);
      }
      timeRecord.get(tag.name).push({
        name: tag.name,
        articleCount: tag.article_count,
        reach: Number(tag.reach),
        date: new Date(tag.year + '-' + tag.month.toString().padStart(2, '0') + '-01'),
      });
    }
    return timeRecord;
  }
}
